@@ -1,4 +1,4 @@
-/*  $Id: messages.pl,v 1.90 2007/06/27 19:52:39 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -29,7 +29,7 @@
     the GNU General Public License.
 */
 
-:- module($messages,
+:- module('$messages',
 	  [ print_message/2,		% +Kind, +Term
 	    print_message_lines/3,	% +Stream, +Prefix, +Lines
 	    message_to_string/2		% +Term, -String
@@ -42,31 +42,49 @@
 :- discontiguous
 	prolog_message/3.
 
+%%	translate_message(+Term)// is det.
+%
+%	Translate a message Term into message lines. The produced lines
+%	is a list of
+%	
+%	    * nl
+%	    Emit a newline
+%	    * Fmt-Args
+%	    Emit the result of format(Fmt, Args)
+%	    * Fmt
+%	    Emit the result of format(Fmt)
+%	    * flush
+%	    Used only as last element of the list.   Simply flush the
+%	    output instead of producing a final newline.
+
 translate_message(Term) -->
+	translate_message2(Term), !.
+
+translate_message2(Term) -->
 	{var(Term)}, !,
 	[ 'Unknown message: ~p'-[Term] ].
-translate_message(Term) -->
+translate_message2(Term) -->
 	prolog:message(Term).
-translate_message(Term) -->
+translate_message2(Term) -->
 	prolog_message(Term).
-translate_message(error(resource_error(stack), Name)) -->
+translate_message2(error(resource_error(stack), Name)) -->
 	[ 'Out of ~w stack'-[Name] ].
-translate_message(error(resource_error(Missing), _)) -->
+translate_message2(error(resource_error(Missing), _)) -->
 	[ 'Not enough resources: ~w'-[Missing] ].
-translate_message(error(ISO, SWI)) -->
+translate_message2(error(ISO, SWI)) -->
 	swi_context(SWI),
 	term_message(ISO),
 	swi_extra(SWI).
-translate_message($aborted) -->
+translate_message2('$aborted') -->
 	[ 'Execution Aborted' ].
-translate_message(message_lines(Lines), L, T) :- % deal with old C-warning()
+translate_message2(message_lines(Lines), L, T) :- % deal with old C-warning()
 	make_message_lines(Lines, L, T).
-translate_message(format(Fmt, Args)) -->
+translate_message2(format(Fmt, Args)) -->
 	[ Fmt-Args ].
-translate_message(Term) -->
+translate_message2(Term) -->
 	{ Term = error(_, _) },
 	[ 'Unknown exception: ~p'-[Term] ].
-translate_message(Term) -->
+translate_message2(Term) -->
 	[ 'Unknown message: ~p'-[Term] ].
 
 make_message_lines([], T, T) :- !.
@@ -97,6 +115,8 @@ iso_message(domain_error(Domain, Actual)) -->
 	[ 'Domain error: `~w'' expected, found `~p'''-[Domain, Actual] ].
 iso_message(instantiation_error) -->
 	[ 'Arguments are not sufficiently instantiated' ].
+iso_message(representation_error(variable)) -->
+	[ 'Cannot represent: argument must be unbound' ].
 iso_message(representation_error(What)) -->
 	[ 'Cannot represent due to `~w'''-[What] ].
 iso_message(permission_error(Action, built_in_procedure, Pred)) -->
@@ -128,6 +148,8 @@ iso_message(busy(Type, Object)) -->
 iso_message(syntax_error(Id)) -->
 	[ 'Syntax error: ' ],
 	syntax_error(Id).
+iso_message(occurs_check(Var, In)) -->
+	[ 'Cannot unify ~p with ~p: would create an infinite tree'-[Var, In] ].
 
 syntax_error(end_of_clause) -->
 	[ 'Unexpected end of clause' ].
@@ -193,7 +215,7 @@ swi_message(signal(Name, Num)) -->
 swi_message(limit_exceeded(Limit, MaxVal)) -->
 	[ 'Exceeded ~w limit (~w)'-[Limit, MaxVal] ].
 swi_message(goal_failed(Goal)) -->
-	[ 'goal (~p) failed'-[Goal] ].
+	[ 'goal unexpectedly failed: ~p'-[Goal] ].
 swi_message(shared_object(_Action, Message)) --> % Message = dlerror() 
 	[ '~w'-[Message] ].
 swi_message(system_error(Error)) -->
@@ -211,6 +233,13 @@ swi_message(not_implemented(Type, What)) -->
 swi_message(context_error(nodirective, Goal)) -->
 	{ goal_to_predicate_indicator(Goal, PI) },
 	[ 'Wrong context: ~p can only be used in a directive'-[PI] ].
+swi_message(context_error(edit, no_default_file)) -->
+	(   { current_prolog_flag(windows, true) }
+	->  [ 'Edit/0 can only be used after opening a \
+	      Prolog file by double-clicking it' ]
+	;   [ 'Edit/0 can only be used with the "-s file" commandline option'
+	    ]
+	).
 swi_message(format_argument_type(Fmt, Arg)) -->
 	[ 'Illegal argument to format sequence ~~~w: ~p'-[Fmt, Arg] ].
 swi_message(format(Msg)) -->
@@ -224,6 +253,8 @@ swi_context(context(Caller, _Msg)) -->
 	{ ground(Caller)
 	}, !,
 	caller(Caller).
+swi_context(file(Path, Line, -1, _CharNo)) --> !,
+	[ '~w:~d: '-[Path, Line] ].
 swi_context(file(Path, Line, LinePos, _CharNo)) -->
 	[ '~w:~d:~d: '-[Path, Line, LinePos] ].
 swi_context(stream(Stream, Line, LinePos, _CharNo)) -->
@@ -231,6 +262,8 @@ swi_context(stream(Stream, Line, LinePos, _CharNo)) -->
 swi_context(_) -->
 	[].
 
+caller(system:'$record_clause'/3) --> !,
+	[].
 caller(Module:Name/Arity) --> !,
 	(   { \+ hidden_module(Module) }
 	->  [ '~q:~q/~w: '-[Module, Name, Arity] ]
@@ -302,6 +335,8 @@ prolog_message(io_warning(Stream, Message)) -->
 	[ '~p:~d:~d: ~w'-[Obj, LineNo, LinePos, Message] ].
 prolog_message(io_warning(Stream, Message)) -->
 	[ 'stream ~p: ~w'-[Stream, Message] ].
+prolog_message(option_usage(pldoc)) -->
+	[ 'Usage: --pldoc[=port]' ].
 
 
 		 /*******************************
@@ -350,7 +385,7 @@ prolog_message(dwim_undefined(Goal, Alternatives)) -->
 	],
 	dwim_message(Alternatives).
 prolog_message(dwim_correct(Into)) -->
-	[ 'Correct to: ~w? '-[Into], flush ].
+	[ 'Correct to: ~q? '-[Into], flush ].
 prolog_message(error(loop_error(Spec), file_search(Used))) -->
 	[ 'File search: too many levels of indirections on: ~p'-[Spec], nl,
 	  '    Used alias expansions:', nl
@@ -467,6 +502,10 @@ prolog_message(version) -->
 	  Patch is Version mod 100
 	},
 	[ '~w.~w.~w'-[Major, Minor, Patch] ].
+prolog_message(address_bits) -->
+	{ current_prolog_flag(address_bits, Bits)
+	}, !,
+	[ '~d bits, '-[Bits] ].
 prolog_message(threads) -->
 	{ current_prolog_flag(threads, true)
 	}, !,
@@ -474,7 +513,7 @@ prolog_message(threads) -->
 prolog_message(threads) -->
 	[].
 prolog_message(copyright) -->
-	[ 'Copyright (c) 1990-2007 University of Amsterdam.', nl,
+	[ 'Copyright (c) 1990-2008 University of Amsterdam.', nl,
 	  'SWI-Prolog comes with ABSOLUTELY NO WARRANTY. This is free software,', nl,
 	  'and you are welcome to redistribute it under certain conditions.', nl,
 	  'Please visit http://www.swi-prolog.org for details.'
@@ -484,6 +523,7 @@ prolog_message(author) -->
 prolog_message(welcome) -->
 	[ 'Welcome to SWI-Prolog (' ],
 	prolog_message(threads),
+	prolog_message(address_bits),
 	['Version ' ],
 	prolog_message(version),
 	[ ')', nl ],
@@ -512,17 +552,62 @@ prolog_message(var_query(_)) -->
 prolog_message(close_on_abort(Stream)) -->
 	[ 'Abort: closed stream ~p'-[Stream] ].
 
-prolog_message(query(no)) -->
-	[ nl, 'No' ].
-prolog_message(query(yes)) -->
-	[ nl, 'Yes' ].
-prolog_message(query(yes, [])) -->
-	[ 'More? ', flush ].
-prolog_message(query(yes, Bindings)) -->
+prolog_message(query(QueryResult)) -->
+	query_result(QueryResult).
+
+query_result(no) -->		% failure
+	[ 'fail.' ],
+	extra_line.
+query_result(yes([])) --> !,	% prompt_alternatives_on: groundness
+	[ 'true.' ],
+	extra_line.
+query_result(yes(Residuals)) -->
+	residuals(Residuals),
+	extra_line.
+query_result(done) -->		% user typed <CR>
+	extra_line.
+query_result(yes(Bindings, Residuals)) -->
 	bindings(Bindings),
-	[ ' ', flush ].
-prolog_message(query(eof)) -->
+	bind_res_sep(Bindings, Residuals),
+	residuals(Residuals),
+	prompt(yes, Bindings, Residuals).
+query_result(more(Bindings, Residuals)) -->
+	bindings(Bindings),
+	bind_res_sep(Bindings, Residuals),
+	residuals(Residuals),
+	prompt(more, Bindings, Residuals).
+query_result(help) -->
+	[ nl, 'Actions:', nl, nl,
+	  '; (n, r, space, TAB): redo    t:          trace & redo', nl,
+	  'b:                    break   c (a, RET): exit', nl,
+	  'w:                    write   p           print', nl,
+	  'h (?):                help',
+	  nl, nl
+	].
+query_result(action) -->
+	[ 'Action? ', flush ].
+query_result(confirm) -->
+	[ 'Please answer \'y\' or \'n\'? ', flush ].
+query_result(eof) -->
 	[ nl ].
+query_result(toplevel_open_line) -->
+	[].
+
+prompt(Answer, [], []) --> !,
+	prompt(Answer, empty).
+prompt(Answer, _, _) --> !,
+	prompt(Answer, non_empty).
+
+prompt(yes, empty) --> !,
+	[ 'true.' ],
+	extra_line.
+prompt(yes, _) --> !,
+	[ '.' ],
+	extra_line.
+prompt(more, empty) --> !,
+	[ 'true ', flush ].
+prompt(more, _) --> !,
+	[ ' ', flush ].
 
 bindings([]) -->
 	[].
@@ -533,18 +618,34 @@ bindings([Name = Value|T]) -->
 	  ;   Cont = (,)
 	  )
 	},
-	[ nl, '~w = ~W~w'-[Name, Value, Options, Cont] ],
+	[ '~w = ~W~w'-[Name, Value, Options, Cont] ],
+	(   {T == []}
+	->  []
+	;   [nl]
+	),
 	bindings(T).
 
-prolog_message(query(help)) -->
-	[ nl, 'Actions:', nl, nl,
-	  '; (n, r):     redo    t:                 trace & redo', nl,
-	  'b:            break   c (a, RET, space): continue', nl,
-	  'w:            write   p                  print', nl,
-	  'h (?):        help', nl, nl
-	].
-prolog_message(query(action)) -->
-	[ 'Action? ', flush ].
+residuals([]) --> !, [].
+residuals([G|Gs]) -->
+	(   { Gs \== [] }
+	->  [ '~q,'-[G], nl ],
+	    residuals(Gs)
+	;   [ '~q'-[G] ]
+	).
+	
+bind_res_sep(_, []) --> !,
+	[].
+bind_res_sep([], _) --> !,
+	[].
+bind_res_sep(_, _) -->
+	[','-[], nl].
+
+extra_line -->
+	{ current_prolog_flag(toplevel_extra_white_line, true) }, !,
+	['~N'-[]].
+extra_line -->
+	[].
+
 prolog_message(if_tty(Text)) -->
 	(   {current_prolog_flag(tty_control, true)}
 	->  [ at_same_line, '~w'-[Text] ]
@@ -554,8 +655,6 @@ prolog_message(halt(Reason)) -->
 	[ '~w: halt'-[Reason] ].
 prolog_message(no_action(Char)) -->
 	[ 'Unknown action: ~c (h for help)'-[Char], nl ].
-prolog_message(query(confirm)) -->
-	[ 'Please answer \'y\' or \'n\'? ', flush ].
 
 prolog_message(history(help(Show, Help))) -->
 	[ 'History Commands:', nl, 
@@ -797,8 +896,8 @@ print_system_message(_, Level, Lines) :-
 	prefix(Level, LinePrefix, Stream), !,
 	print_message_lines(Stream, LinePrefix, Lines).
 	
-prefix(error,	      'ERROR: (~w:~d):~n',   '\t', '', 0.5, user_error).
-prefix(warning,	      'Warning: (~w:~d):~n', '\t', '', 0,   user_error).
+prefix(error,	      'ERROR: ~w:~d:~n',   '\t', '', 0.5, user_error).
+prefix(warning,	      'Warning: ~w:~d:~n', '\t', '', 0,   user_error).
 
 prefix(help,	      '',          user_error).
 prefix(query,	      '',          user_error).
@@ -865,7 +964,7 @@ actions_to_format([nl|T], Fmt, Args) :- !,
 actions_to_format([Fmt0-Args0|Tail], Fmt, Args) :- !,
         actions_to_format(Tail, Fmt1, Args1),
         atom_concat(Fmt0, Fmt1, Fmt),
-        $append(Args0, Args1, Args).
+        '$append'(Args0, Args1, Args).
 actions_to_format([Term|Tail], Fmt, Args) :-
 	atomic(Term), !,
         actions_to_format(Tail, Fmt1, Args),
@@ -873,7 +972,7 @@ actions_to_format([Term|Tail], Fmt, Args) :-
 actions_to_format([Term|Tail], Fmt, Args) :-
         actions_to_format(Tail, Fmt1, Args1),
         atom_concat('~w', Fmt1, Fmt),
-        $append([Term], Args1, Args).
+        '$append'([Term], Args1, Args).
 
 
 		 /*******************************
