@@ -1,4 +1,4 @@
-/*  $Id: toplevel.pl,v 1.97 2006/08/16 08:53:29 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -29,7 +29,7 @@
     the GNU General Public License.
 */
 
-:- module($toplevel,
+:- module('$toplevel',
 	[ '$initialise'/0		% start Prolog (does not return)
 	, '$toplevel'/0			% Prolog top-level (re-entrant)
 	, '$abort'/0 			% restart after an abort
@@ -227,7 +227,7 @@ path_sep -->
 		 *   LOADING ASSIOCIATED FILES	*
 		 *******************************/
 
-%	set_associated_file/0
+%%	set_associated_file
 %	
 %	If SWI-Prolog is started as <exe> <file>.<ext>, where <ext> is
 %	the extension registered for associated files, set the Prolog
@@ -242,24 +242,49 @@ set_associated_file :-
 	current_prolog_flag(argv, Argv),
 	'$append'(Pre, [OsFile], Argv),
 	\+ memberchk(--, Pre),
+	\+ '$append'(_, ['-f'], Pre),	% Avoid loading twice
 	prolog_to_os_filename(File, OsFile),
 	file_name_extension(_, Ext, File),
 	access_file(File, read), !,
 	file_directory_name(File, Dir),
 	working_directory(_, Dir),
 	set_prolog_flag(associated_file, File),
-	atom_concat('SWI-Prolog -- ', File, Title),
-	(   '$c_current_predicate'(_, system:window_title(_, _))
-	->  system:window_title(_, Title)
+	(   current_predicate(system:window_title/2)
+	->  atom_concat('SWI-Prolog -- ', File, Title),
+	    system:window_title(_, Title)
 	;   true
 	).
 set_associated_file.
 
 
-%	load_associated_file/0
+%%	start_pldoc
+%
+%	If the option  =|--pldoc[=port]|=  is   given,  load  the PldDoc
+%	system.
+
+start_pldoc :-
+	current_prolog_flag(argv, Argv),
+	'$member'(Av, Argv),
+	(   Av == (--)
+	->  !
+	;   atom_concat('--pldoc', Rest, Av)
+	->  (   Rest == ''
+	    ->	call((doc_server(_),
+		      doc_browser))
+	    ;	atom_concat(=, PortAtom, Rest),
+		catch(atom_number(PortAtom, Port), _, fail)
+	    ->	call(doc_server(Port))
+	    ;	print_message(error, option_usage(pldoc)),
+		halt(1)
+	    )
+	).
+start_pldoc.
+
+
+%%	load_associated_file
 %	
-%	Load the file-name set by set_associated_file/0 from the
-%	commandline arguments. Not the expand(false) to avoid expanding
+%	Load  the  file-name  set  by   set_associated_file/0  from  the
+%	commandline arguments. Note the expand(false) to avoid expanding
 %	special characters in the filename.
 
 load_associated_file :-
@@ -304,13 +329,16 @@ initialise_prolog :-
 	'$set_file_search_paths',
 	once(print_predicate(_, [print], PrintOptions)),
 	set_prolog_flag(toplevel_print_options, PrintOptions),
+	set_prolog_flag(prompt_alternatives_on, determinism),
+	set_prolog_flag(toplevel_extra_white_line, true),
 	'$set_debugger_print_options'(print),
 	'$run_at_initialization',
 	'$load_system_init_file',
 	'$load_gnu_emacs_interface',
 	'$option'(init_file, OsFile, OsFile),
 	prolog_to_os_filename(File, OsFile),
-	'$load_init_file'(File), 
+	'$load_init_file'(File),
+	start_pldoc,
 	'$load_script_file',
 	load_associated_file,
 	'$option'(goal, GoalAtom, GoalAtom), 
@@ -551,37 +579,77 @@ subst_chars([H|T]) -->
 	print_message(query, query(no)),
 	fail.
 
-%	write_bindings(+Bindings, +Deterministic)
+%%	write_bindings(+Bindings, +Deterministic)
 %	
-%	Write    bindings    resulting     from      a     query.     If
-%	prompt_alternatives_no_bindings  is  true  we  also  prompt  for
-%	alternatives  if  the  query  makes  no  bindings  but  succeeds
-%	non-deterministically, so the user can   prompt  for alternative
-%	side-effects.
+%	Write   bindings   resulting   from   a     query.    The   flag
+%	prompt_alternatives_on determines whether the   user is prompted
+%	for alternatives. =groundness= gives   the  classical behaviour,
+%	=determinism= is considered more adequate and informative.
 
-write_bindings(Bindings0, Det) :-
-	bind_vars(Bindings0),
-	filter_bindings(Bindings0, Bindings),
-	write_bindings2(Bindings, Det).
+write_bindings(Bindings, Det) :-
+	'$attributed'(Bindings), !,
+	copy_term(Bindings, Bindings1, Residuals0),
+	'$module'(TypeIn, TypeIn),
+	omit_qualifiers(Residuals0, TypeIn, Residuals),
+	bind_vars(Bindings1),
+	filter_bindings(Bindings1, Bindings2),
+	write_bindings2(Bindings2, Residuals, Det).
+write_bindings(Bindings, Det) :-
+	bind_vars(Bindings),
+	filter_bindings(Bindings, Bindings1),
+	write_bindings2(Bindings1, [], Det).
 
-write_bindings2([], Det) :-
-	(   Det == true
-	;   \+ current_prolog_flag(prompt_alternatives_no_bindings, true)
-	), !,
-	print_message(query, query(yes)).
-write_bindings2(Bindings, _Det) :-
+write_bindings2([], Residuals, _) :-
+	current_prolog_flag(prompt_alternatives_on, groundness), !,
+	print_message(query, query(yes, Residuals)).
+write_bindings2(Bindings, Residuals, true) :-
+	current_prolog_flag(prompt_alternatives_on, determinism), !,
+	print_message(query, query(yes(Bindings, Residuals))).
+write_bindings2(Bindings, Residuals, _Det) :-
 	repeat,
-	    print_message(query, query(yes, Bindings)),
+	    print_message(query, query(more(Bindings, Residuals))),
 	    get_respons(Action),
 	(   Action == redo
 	->  !, fail
 	;   Action == show_again
 	->  fail
 	;   !,
-	    print_message(query, query(yes))
+	    print_message(query, query(done))
 	).
 
-%	bind_vars(+Bindings)
+
+%%	omit_qualifiers(+QGoals, +TypeIn, -Goals) is det.
+%
+%	Omit unneeded module qualifiers  from   QGoals  relative  to the
+%	given module TypeIn.
+
+omit_qualifiers([], _, []).
+omit_qualifiers([Goal0|Goals0], TypeIn, [Goal|Goals]) :-
+	omit_qualifier(Goal0, TypeIn, Goal),
+	omit_qualifiers(Goals0, TypeIn, Goals).
+
+omit_qualifier(M:G0, TypeIn, G) :-
+	predicate_property(TypeIn:G0, imported_from(M)), !,
+	omit_meta_qualifiers(G0, TypeIn, G).
+omit_qualifier(_:G0, TypeIn, G) :-
+	predicate_property(G0, built_in), !,
+	omit_meta_qualifiers(G0, TypeIn, G).
+omit_qualifier(M:G0, _, M:G) :-
+	atom(M), !,
+	omit_meta_qualifiers(G0, M, G).
+omit_qualifier(G0, TypeIn, G) :-
+	omit_meta_qualifiers(G0, TypeIn, G).
+
+omit_meta_qualifiers(freeze(V, QGoal), TypeIn, freeze(V, Goal)) :-
+	callable(QGoal), !,
+	omit_qualifier(QGoal, TypeIn, Goal).
+omit_meta_qualifiers(when(Cond, QGoal), TypeIn, when(Cond, Goal)) :-
+	callable(QGoal), !,
+	omit_qualifier(QGoal, TypeIn, Goal).
+omit_meta_qualifiers(G, _, G).
+
+
+%%	bind_vars(+Bindings)
 %	
 %	Bind variables to '$VAR'(Name), so they are printed by the names
 %	used in the query. Note that by   binding  in the reverse order,
@@ -595,7 +663,7 @@ bind_vars([Name=Var|T]) :-
 	;   true
 	).
 
-%	filter_bindings(+Bindings0, -Bindings)
+%%	filter_bindings(+Bindings0, -Bindings)
 %	
 %	Remove bindings that must not be printed.
 
@@ -627,14 +695,15 @@ answer_respons(Char, again) :-
 	memberchk(Char, "?h"), !,
 	print_message(help, query(help)).
 answer_respons(Char, redo) :-
-	memberchk(Char, ";nrNR"), !,
+	memberchk(Char, ";nrNR \t"), !,
 	print_message(query, if_tty(';')).
 answer_respons(Char, redo) :-
 	memberchk(Char, "tT"), !,
 	trace,
 	print_message(query, if_tty('; [trace]')).
 answer_respons(Char, continue) :-
-	memberchk(Char, [0'c, 0'a, 0' , 10, 13, 0'y, 0'Y]), !.
+	memberchk(Char, "ca\n\ryY."), !,
+	print_message(query, if_tty('.')).
 answer_respons(0'b, show_again) :- !,
 	break.
 answer_respons(Char, show_again) :-
@@ -648,12 +717,12 @@ answer_respons(Char, again) :-
 	print_message(query, no_action(Char)).
 
 print_predicate(0'w, [write], [ quoted(true),
-				attributes(write)
+				priority(699)
 			      ]).
 print_predicate(0'p, [print], [ quoted(true),
 				portray(true),
-				attributes(portray),
-				max_depth(10)
+				max_depth(10),
+				priority(699)
 			      ]).
 
 
