@@ -1,11 +1,11 @@
-/*  $Id: thread_httpd.pl,v 1.16 2007/03/09 14:47:37 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -33,9 +33,11 @@
 	  [ http_current_server/2,	% ?:Goal, ?Port
 	    http_server/2,		% :Goal, +Options
 	    http_workers/2,		% +Port, ?WorkerCount
-	    http_current_worker/2	% ?Port, ?ThreadID
+	    http_current_worker/2,	% ?Port, ?ThreadID
+	    http_stop_server/2		% +Port, +Options
 	  ]).
 :- use_module(library(debug)).
+:- use_module(library(error)).
 :- use_module(library(option)).
 :- use_module(library(lists)).
 :- use_module(library(socket)).
@@ -59,13 +61,14 @@ for details.
 	http_current_server(:, ?).
 
 :- dynamic
-	current_server/3,		% Port, Goal, Queue
+	current_server/4,		% Port, Goal, Thread, Queue
 	queue_worker/2,			% Queue, ThreadID
 	queue_options/2.		% Queue, Options
 
 :- multifile
 	make_socket_hook/3,
 	accept_hook/2,
+	close_hook/1,
 	open_client_hook/5.
 
 %%	http_server(:Goal, +Options) is det.
@@ -84,7 +87,7 @@ for details.
 
 http_server(Goal, Options) :-
 	strip_module(Goal, Module, G),
-	select(port(Port), Options, Options1), !,
+	select_option(port(Port), Options, Options1), !,
 	after_option(Options1, Module, Options2),
 	make_socket(Port, Options2, Options3),
 	create_workers(Options3),
@@ -126,13 +129,12 @@ create_server(Goal, Port, Options) :-
 	memberchk(queue(Queue), Options),
 	atom_concat('http@', Port, Alias),
 	thread_create(accept_server(Goal, Options), _,
-		      [ detached(true),
-			local(128),
+		      [ local(128),
 			global(128),
 			trail(128),
 			alias(Alias)
 		      ]),
-	assert(current_server(Port, Goal, Queue)).
+	assert(current_server(Port, Goal, Alias, Queue)).
 
 
 %%	http_current_server(:Goal, ?Port) is nondet.
@@ -140,8 +142,11 @@ create_server(Goal, Port, Options) :-
 %	True if Goal is the goal of a server at Port.
 
 http_current_server(Goal, Port) :-
-	strip_module(Goal, Module, G),
-	current_server(Port, Module:G, _).
+	(   var(Goal)
+	->  current_server(Port, Goal, _, _)
+	;   strip_module(Goal, Module, G),
+	    current_server(Port, Module:G, _, _)
+	).
 
 
 %%	http_workers(+Port, -Workers) is det.
@@ -152,15 +157,18 @@ http_current_server(Goal, Port) :-
 %	(one) can be used to profile the worker using tprofile/1.
 
 http_workers(Port, Workers) :-
-	current_server(Port, _, Queue),
+	must_be(integer, Port),
+	current_server(Port, _, _, Queue), !,
 	(   integer(Workers)
 	->  resize_pool(Queue, Workers)
 	;   findall(W, queue_worker(Queue, W), WorkerIDs),
 	    length(WorkerIDs, Workers)
 	).
+http_workers(Port, _) :-
+	existence_error(http_server, Port).
 
 
-%%	http_current_worker(?Port, ?ThreadID)
+%%	http_current_worker(?Port, ?ThreadID) is nondet.
 %	
 %	True if ThreadID is the identifier   of  a Prolog thread serving
 %	Port. This predicate is  motivated  to   allow  for  the  use of
@@ -168,7 +176,7 @@ http_workers(Port, Workers) :-
 %	statistics.
 
 http_current_worker(Port, ThreadID) :-
-	current_server(Port, _, Queue),
+	current_server(Port, _, _, Queue),
 	queue_worker(Queue, ThreadID).
 
 
@@ -178,18 +186,62 @@ http_current_worker(Port, ThreadID) :-
 %	posting them to the queue of workers.
 
 accept_server(Goal, Options) :-
+	catch(accept_server2(Goal, Options), http_stop, true),
+	thread_self(Thread),
+	retract(current_server(_Port, _, Thread, _Queue)),
+	close_server_socket(Options).
+
+accept_server2(Goal, Options) :-
 	accept_hook(Goal, Options), !.
-accept_server(Goal, Options) :-
+accept_server2(Goal, Options) :-
 	memberchk(tcp_socket(Socket), Options), !,
 	memberchk(queue(Queue), Options),
 	repeat,
-	  catch(tcp_accept(Socket, Client, Peer), E,
-		(   print_message(error, E),
-		    fail
-		)),
-	  debug(http(connection), 'New HTTP connection from ~p', [Peer]),
-	  thread_send_message(Queue, tcp_client(Client, Goal, Peer)),
-	fail.
+	  (   catch(tcp_accept(Socket, Client, Peer), E, true)
+	  ->  (   var(E)
+	      ->  debug(http(connection), 'New HTTP connection from ~p', [Peer]),
+		  thread_send_message(Queue, tcp_client(Client, Goal, Peer)),
+		  fail
+	      ;   E == http_stop
+	      ->  throw(http_stop)
+	      ;   print_message(error, E),
+		  fail
+	      )
+	  ;   print_message(error, goal_failed(tcp_accept(Socket, Client, Peer)))
+	  ).
+
+
+%%	close_server_socket(+Options)
+%
+%	Close the server socket.
+
+close_server_socket(Options) :-
+	close_hook(Options), !.
+close_server_socket(Options) :-
+	memberchk(tcp_socket(Socket), Options), !,
+	tcp_close_socket(Socket).
+
+
+%%	http_stop_server(+Port, +Options)
+%
+%	Stop the indicated  HTTP  server   gracefully.  First  stops all
+%	workers, then stops the server.
+%	
+%	@tbd	Realise non-graceful stop
+
+http_stop_server(Port, _Options) :-
+	http_workers(Port, 0),
+	current_server(Port, _, Thread, Queue),
+	retractall(queue_options(Queue, _)),
+	thread_signal(Thread, throw(http_stop)),
+	catch(connect(localhost:Port), _, true),
+	thread_join(Thread, _),
+	message_queue_destroy(Queue).
+
+connect(Address) :-
+        tcp_socket(Socket),
+        tcp_connect(Socket, Address),
+	tcp_close_socket(Socket).
 
 
 		 /*******************************
@@ -222,6 +274,11 @@ create_workers(I, N, Queue, AliasBase, Options) :-
 	create_workers(I2, N, Queue, AliasBase, Options).
 
 
+%%	resize_pool(+Queue, +Workers) is det.
+%
+%	Create or destroy workers. If workers   are  destroyed, the call
+%	waits until the desired number of waiters is reached.
+
 resize_pool(Queue, Size) :-
 	findall(W, queue_worker(Queue, W), Workers),
 	length(Workers, Now),
@@ -234,14 +291,23 @@ resize_pool(Queue, Size) :-
 	->  true
 	;   Now > Size
 	->  Excess is Now - Size,
-	    forall(between(1, Excess, _), thread_send_message(Queue, quit))
+	    thread_self(Me),
+	    forall(between(1, Excess, _), thread_send_message(Queue, quit(Me))),
+	    forall(between(1, Excess, _), thread_get_message(quitted(_)))
 	).
 
 
 %%	http_worker(+Options)
 %	
-%	A worker.  Workers simply wait until they are passes an accepted
-%	socket to process a client.
+%	Run HTTP worker main loop. Workers   simply  wait until they are
+%	passed an accepted socket to process  a client. After processing
+%	a request they keep processing requests from the same connection
+%	as long as =|Connection:  keep-alive|=   remains  requested from
+%	both parties. Otherwise they close the  connection and return to
+%	the worker pool.  See server_loop/6.
+%	
+%	If the message quit(Sender) is read   from the queue, the worker
+%	stops.
 
 http_worker(Options) :-
 	option(timeout(Timeout), Options, infinite),
@@ -249,9 +315,10 @@ http_worker(Options) :-
 	thread_at_exit(done_worker),
 	repeat,
 	  thread_get_message(Queue, Message),
-	  (   Message == quit
+	  (   Message = quit(Sender)
 	  ->  thread_self(Self),
-	      thread_detach(Self)
+	      thread_detach(Self),
+	      thread_send_message(Sender, quitted(Self))
 	  ;   open_client(Message, Goal, In, Out, ClientOptions),
 	      set_stream(In, timeout(Timeout)),
 	      debug(http(server), 'Running server goal ~p on ~p -> ~p',
@@ -297,7 +364,7 @@ open_client(tcp_client(Socket, Goal, Peer), Goal, In, Out,
 
 done_worker :-
 	thread_self(Self),
-	current_thread(Self, Status),
+	thread_property(Self, status(Status)),
 	retract(queue_worker(_Queue, Self)),
 	print_message(informational,
 		      httpd_stopped_worker(Self, Status)).
@@ -384,3 +451,4 @@ prolog:message(httpd_stopped_worker(Self, Status)) -->
 	prolog:meta_goal/2.
 
 prolog:meta_goal(http_server(G, _), [G+1]).
+prolog:meta_goal(http_current_server(G, _), [G+1]).
