@@ -1,11 +1,11 @@
-/*  $Id: plunit.pl,v 1.21 2007/03/28 08:51:25 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2006, University of Amsterdam
+    Copyright (C): 2006-2007, University of Amsterdam
 
     This file is covered by the `The Artistic License', also in use by
     Perl.  See http://www.perl.com/pub/a/language/misc/Artistic.html
@@ -40,38 +40,51 @@ please visit http://www.swi-prolog.org/packages/plunit.html.
 :- dynamic
 	include_code/1.
 
+including :-
+	include_code(X), !,
+	X == true.
+including.
+
 if_expansion((:- if(G)), []) :-
-	(   catch(G, E, (print_message(error, E), fail))
-	->  asserta(include_code(true))
-	;   asserta(include_code(false))
+	(   including
+	->  (   catch(G, E, (print_message(error, E), fail))
+	    ->  asserta(include_code(true))
+	    ;   asserta(include_code(false))
+	    )
+	;   asserta(include_code(else_false))
 	).
 if_expansion((:- else), []) :-
 	(   retract(include_code(X))
 	->  (   X == true
 	    ->  X2 = false 
-	    ;   X2 = true
+	    ;   X == false
+	    ->	X2 = true
+	    ;	X2 = X
 	    ),
 	    asserta(include_code(X2))
-	;   throw(error(context_error(no_if), _))
+	;   throw_error(context_error(no_if),_)
 	).
 if_expansion((:- endif), []) :-
 	retract(include_code(_)), !.
 
 if_expansion(_, []) :-
-	include_code(X), !,
-	X == false.
+	\+ including.
 	    
 user:term_expansion(In, Out) :-
 	prolog_load_context(module, plunit),
 	if_expansion(In, Out).
 
-swi     :- catch(current_prolog_flag(compiled_at, _), _, fail).
+swi     :- catch(current_prolog_flag(dialect, swi), _, fail).
 sicstus :- catch(current_prolog_flag(system_type, _), _, fail).
 
 
 :- if(swi).
+throw_error(Error_term,Impldef) :-
+	throw(error(Error_term,Impldef)).
+
 :- set_prolog_flag(generate_debug_info, false).
 :- use_module(library(option)).
+:- use_module(library(pairs)).
 
 current_test_flag(Name, Value) :-
 	current_prolog_flag(Name, Value).
@@ -81,6 +94,8 @@ set_test_flag(Name, Value) :-
 :- endif.
 
 :- if(sicstus).
+throw_error(Error_term,Impldef) :-
+	throw(error(Error_term,i(Impldef))). % SICStus 3 work around
 :- use_module(swi).			% SWI-Compatibility
 :- use_module(library(terms)).
 :- op(700, xfx, =@=).
@@ -108,7 +123,7 @@ current_test_flag(Name, Val) :-
 
 set_test_flag(Name, Val) :-
 	var(Name), !,
-	throw(error(instantiation_error, set_test_flag(Name,Val))).
+	throw_error(instantiation_error, set_test_flag(Name,Val)).
 set_test_flag( Name, Val ) :-
 	retractall(test_flag(Name,_)),
 	asserta(test_flag(Name, Val)).
@@ -125,7 +140,8 @@ set_test_flag( Name, Val ) :-
    (   current_test_flag(test_options, _)
    ->  true
    ;   set_test_flag(test_options,
-		     [ run(make)	% run tests on make/0
+		     [ run(make),	% run tests on make/0
+		       sto(false)
 		     ])
    ).
 
@@ -141,6 +157,13 @@ set_test_flag( Name, Val ) :-
 %		When the tests are run.  Values are =manual=, =make=
 %		or make(all).
 %		
+%		* sto(+Bool)
+%		How to test whether code is subject to occurs check
+%		(STO).  If =false= (default), STO is not considered.
+%		If =true= and supported by the hosting Prolog, code
+%		is run in all supported unification mode and reported
+%		if the results are inconsistent.
+%		
 %	@tbd	Verify types	
 
 set_test_options(Options) :-
@@ -178,12 +201,14 @@ begin_tests(Unit) :-
 	begin_tests(Unit, []).
 
 begin_tests(Unit, Options) :-
+	valid_options(Options, test_set_option),
 	make_unit_module(Unit, Name),
 	source_location(File, Line),
 	begin_tests(Unit, Name, File:Line, Options).
 
 :- if(swi).
 begin_tests(Unit, Name, File:Line, Options) :-
+	loading_tests, !,
 	'$set_source_module'(Context, Context),
 	Supers = [Context],
 	(   current_unit(Unit, Name, Supers, Options)
@@ -198,6 +223,9 @@ begin_tests(Unit, Name, File:Line, Options) :-
 	'$set_predicate_attribute'(Name:'unit test'/4, trace, 0),
 	discontiguous(Name:'unit body'/2),
 	asserta(loading_unit(Unit, Name, File, Old)).
+begin_tests(Unit, Name, File:_Line, _Options) :-
+	'$set_source_module'(Old, Old),
+	asserta(loading_unit(Unit, Name, File, Old)).
 
 set_import_modules(Module, Imports) :-
 	findall(I, import_module(Module, I), IL),
@@ -210,17 +238,21 @@ set_import_modules(Module, Imports) :-
 
 user:term_expansion((:- begin_tests(Set)),
 		    [ (:- begin_tests(Set)),
+		      (:- discontiguous(test/2)),
 		      (:- discontiguous('unit body'/2)),
 		      (:- discontiguous('unit test'/4))
 		    ]).
 
 begin_tests(Unit, Name, File:_Line, Options) :-
+	loading_tests, !,
 	(   current_unit(Unit, Name, Supers, Options)
 	->  true
 	;   retractall(current_unit(Unit, Name, _, _)),
 	    assert(current_unit(Unit, Name, Supers, Options))
 	),
 	asserta(loading_unit(Unit, Name, File, -)).
+begin_tests(Unit, Name, File:_Line, _Options) :-
+	asserta(loading_unit(Unit, Name, File, -)).
 
 :- endif.
 
@@ -236,10 +268,10 @@ end_tests(Unit) :-
 	(   Unit == StartUnit
 	->  once(retract(loading_unit(StartUnit, _, _, Old))),
 	    '$set_source_module'(_, Old)
-	;   throw(error(context_error(plunit_close(Unit, StartUnit)), _))
+	;   throw_error(context_error(plunit_close(Unit, StartUnit)), _)
 	).
 end_tests(Unit) :-
-	throw(error(context_error(plunit_close(Unit, -)), _)).
+	throw_error(context_error(plunit_close(Unit, -)), _).
 
 %%	make_unit_module(+Name, -ModuleName) is det.
 %%	unit_module(+Name, -ModuleName) is det.
@@ -253,8 +285,8 @@ make_unit_module(Unit, Module) :-
 	unit_module(Unit, Module),
 	(   current_module(Module),
 	    \+ current_unit(_, Module, _, _)
-	->  throw(error(permission_error(create, plunit, Unit),
-			'Existing module'))
+	->  throw_error(permission_error(create, plunit, Unit),
+			'Existing module')
 	;  true
 	).
 
@@ -278,7 +310,8 @@ make_unit_module(Unit, Module) :-
 
 %%	expand_test(+Name, +Options, +Body, -Clause) is det.
 %
-%	@tbd	Verify options.
+%	Expand test(Name, Options) :-  Body  into   a  clause  for
+%	'unit test'/4 and 'unit body'/2.
 
 expand_test(Name, Options0, Body,
 	    [ 'unit test'(Name, Line, Options, Module:'unit body'(Id, Vars)),
@@ -290,9 +323,21 @@ expand_test(Name, Options0, Body,
 	term_variables(Body, VarList),
 	Vars =.. [vars|VarList],
 	(   is_list(Options0)		% allow for single option without list
-	->  Options = Options0
-	;   Options = [Options0]
-	).
+	->  Options1 = Options0
+	;   Options1 = [Options0]
+	),
+	maplist(expand_option, Options1, Options),
+	valid_options(Options, test_option).
+
+expand_option(Var, _) :-
+	var(Var), !,
+	throw_error(instantiation_error,_).
+expand_option(A == B, true(A==B)) :- !.
+expand_option(A = B, true(A=B)) :- !.
+expand_option(A =@= B, true(A=@=B)) :- !.
+expand_option(A =:= B, true(A=:=B)) :- !.
+expand_option(O, O).
+
 
 %%	expand(+Term, -Clauses) is semidet.
 
@@ -306,6 +351,10 @@ expand((test(Name) :- Body), Clauses) :- !,
 	expand_test(Name, [], Body, Clauses).
 expand((test(Name, Options) :- Body), Clauses) :- !,
 	expand_test(Name, Options, Body, Clauses).
+expand(test(Name), _) :- !,
+	throw_error(existence_error(body, test(Name)), _).
+expand(test(Name, _Options), _) :- !,
+	throw_error(existence_error(body, test(Name)), _).
 
 :- if(swi).
 :- multifile
@@ -319,6 +368,81 @@ user:term_expansion(Term, Expanded) :-
 	).
 
 
+		 /*******************************
+		 *	       OPTIONS		*
+		 *******************************/
+
+:- if(swi).
+:- use_module(library(error)).
+:- else.
+must_be(list, X) :- !,
+	(   is_list(X)
+	->  true
+	;   is_not(list, X)
+	).
+must_be(Type, X) :-
+	(   call(Type, X)
+	->  true
+	;   is_not(Type, X)
+	).
+
+is_not(Type, X) :-
+	(   ground(X)
+	->  throw_error(type_error(Type, X), _)
+	;   throw_error(instantiation_error, _)
+	).
+:- endif.
+
+%%	valid_options(+Options, :Pred) is det.
+%
+%	Verify Options to be a list of valid options according to
+%	Pred.
+%
+%	@throws	=type_error= or =instantiation_error=.
+
+valid_options(Options, Pred) :-
+	must_be(list, Options),
+	verify_options(Options, Pred).
+
+verify_options([], _).
+verify_options([H|T], Pred) :-
+	(   call(Pred, H)
+	->  verify_options(T, Pred)
+	;   throw_error(domain_error(Pred, H), _)
+	).
+
+
+%%	test_option(+Option) is semidet.
+%
+%	True if Option is a valid option for test(Name, Options).
+
+test_option(Option) :-
+	test_set_option(Option), !.
+test_option(true(_)).
+test_option(fail).
+test_option(true).
+test_option(throws(_)).
+test_option(error(_)).
+test_option(all(_)).
+test_option(set(_)).
+test_option(nondet).
+test_option(sto(V)) :- nonvar(V), member(V, [finite_trees, rational_trees]).
+
+%%	test_option(+Option) is semidet.
+%
+%	True if Option is a valid option for :- begin_tests(Name,
+%	Options).
+
+test_set_option(blocked(X)) :-
+	must_be(ground, X).
+test_set_option(condition(X)) :-
+	must_be(callable, X).
+test_set_option(setup(X)) :-
+	must_be(callable, X).
+test_set_option(cleanup(X)) :-
+	must_be(callable, X).
+
+
 		 /*******************************
 		 *	  RUNNING TOPLEVEL	*
 		 *******************************/
@@ -326,7 +450,8 @@ user:term_expansion(Term, Expanded) :-
 :- dynamic
 	passed/5,			% Unit, Test, Line, Det, Time
 	failed/4,			% Unit, Test, Line, Reason
-	blocked/4.			% Unit, Test, Line, Reason
+	blocked/4,			% Unit, Test, Line, Reason
+	sto/5.				% Unit, Test, Line, Results
 
 %%	run_tests is semidet.
 %%	run_tests(+TestSet) is semidet.
@@ -350,7 +475,7 @@ run_unit(Spec) :-
 	unit_from_spec(Spec, Unit, Tests, Module, UnitOptions),
 	(   option(blocked(Reason), UnitOptions)
 	->  info(plunit(blocked(unit(Unit, Reason))))
-	;   setup(Module, UnitOptions)
+	;   setup(Module, unit(Unit), UnitOptions)
 	->  info(plunit(begin(Spec))),
 	    forall((Module:'unit test'(Name, Line, Options, Body),
 		    matching_test(Name, Tests)),
@@ -366,10 +491,17 @@ run_unit(Spec) :-
 
 unit_from_spec(Unit, Unit, _, Module, Options) :-
 	atom(Unit), !,
-	current_unit(Unit, Module, _Supers, Options).
+	(   current_unit(Unit, Module, _Supers, Options)
+	->  true
+	;   throw_error(existence_error(unit_test, Unit), _)
+	).
 unit_from_spec(Unit:Tests, Unit, Tests, Module, Options) :-
 	atom(Unit), !,
-	current_unit(Unit, Module, _Supers, Options).
+	(   current_unit(Unit, Module, _Supers, Options)
+	->  true
+	;   throw_error(existence_error(unit_test, Unit), _)
+	).
+
 
 matching_test(X, X) :- !.
 matching_test(Name, Set) :-
@@ -379,7 +511,8 @@ matching_test(Name, Set) :-
 cleanup :-
 	retractall(passed(_, _, _, _, _)),
 	retractall(failed(_, _, _, _)),
-	retractall(blocked(_, _, _, _)).
+	retractall(blocked(_, _, _, _)),
+	retractall(sto(_, _, _, _)).
 
 
 %%	run_tests_in_files(+Files:list)	is det.
@@ -388,7 +521,10 @@ cleanup :-
 
 run_tests_in_files(Files) :-
 	findall(Unit, unit_in_files(Files, Unit), Units),
-	run_tests(Units).
+	(   Units == []
+	->  true
+	;   run_tests(Units)
+	).
 
 unit_in_files(Files, Unit) :-
 	is_list(Files), !,
@@ -419,122 +555,245 @@ make_run_tests(Files) :-
 	;   true
 	).
 
+:- if(swi).
+
+unification_capability(sto_error_incomplete).
+% can detect some (almost all) STO runs
+unification_capability(rational_trees).
+unification_capability(finite_trees).
+
+set_unification_capability(Cap) :-
+	cap_to_flag(Cap, Flag),
+	set_prolog_flag(occurs_check, Flag).
+
+current_unification_capability(Cap) :-
+	current_prolog_flag(occurs_check, Flag),
+	cap_to_flag(Cap, Flag), !.
+
+cap_to_flag(sto_error_incomplete, error).
+cap_to_flag(rational_trees, false).
+cap_to_flag(finite_trees, true).
+
+:- else.
+:- if(sicstus).
+
+unification_capability(rational_trees).
+set_unification_capability(rational_trees).
+current_unification_capability(rational_trees).
+
+:- else.
+
+unification_capability(_) :-
+	fail.
+
+:- endif.
+:- endif.
+
 
 		 /*******************************
 		 *	   RUNNING A TEST	*
 		 *******************************/
 
 %%	run_test(+Unit, +Name, +Line, +Options, +Body) is det.
+%
+%	Run a single test.  
 
-run_test(Unit, Name, Line, Options, _Body) :-
-	option(blocked(Reason), Options), !,
+run_test(Unit, Name, Line, Options, Body) :-
+	current_test_flag(test_options, Options),
+	option(sto(false), Options, false), !,
+	run_test_6(Unit, Name, Line, Options, Body, Result),
+	report_result(Result, Options).
+run_test(Unit, Name, Line, Options, Body) :-
+	current_unification_capability(Cap0),
+	call_cleanup(run_test_cap(Unit, Name, Line, Options, Body),
+		     set_unification_capability(Cap0)).
+
+run_test_cap(Unit, Name, Line, Options, Body) :-
+	(   option(sto(Type), Options)
+	->  unification_capability(Type),
+	    set_unification_capability(Type),
+	    run_test_6(Unit, Name, Line, Options, Body, Result),
+	    report_result(Result, Options)
+	;   findall(Key-(Type+Result),
+		    test_caps(Type, Unit, Name, Line, Options, Body, Result, Key),
+		    Pairs),
+	    group_pairs_by_key(Pairs, Keyed),
+	    (	Keyed == []
+	    ->	true
+	    ;	Keyed = [_-Results]
+	    ->	Results = [_Type+Result|_],
+		report_result(Result, Options)		% consistent results
+	    ;	pairs_values(Pairs, ResultByType),
+		report_result(sto(Unit, Name, Line, ResultByType), Options)
+	    )
+	).
+
+%%	test_caps(-Type, +Unit, +Name, +Line, +Options, +Body, -Result, -Key) is nondet.
+
+test_caps(Type, Unit, Name, Line, Options, Body, Result, Key) :-
+	unification_capability(Type),
+	set_unification_capability(Type),
+	run_test_6(Unit, Name, Line, Options, Body, Result),
+	result_to_key(Result, Key),
+	Key \== setup_failed.
+
+result_to_key(blocked(_, _, _, _), blocked).
+result_to_key(failure(_, _, _, How0), failure(How1)) :-
+	( How0 = succeeded(_T) -> How1 = succeeded ; How0 = How1 ).
+result_to_key(success(_, _, _, Determinism, _), success(Determinism)).
+result_to_key(setup_failed(_,_,_), setup_failed).
+
+report_result(blocked(Unit, Name, Line, Reason), _) :- !,
 	assert(blocked(Unit, Name, Line, Reason)).
-run_test(Unit, Name, Line, Options, Body) :-
+report_result(failure(Unit, Name, Line, How), Options) :- !,
+	failure(Unit, Name, Line, How, Options).
+report_result(success(Unit, Name, Line, Determinism, Time), Options) :- !,
+	success(Unit, Name, Line, Determinism, Time, Options).
+report_result(setup_failed(_Unit, _Name, _Line), _Options).
+report_result(sto(Unit, Name, Line, ResultByType), Options) :-
+	assert(sto(Unit, Name, Line, ResultByType)),
+	print_message(error, plunit(sto(Unit, Name, Line))),
+	report_sto_results(ResultByType, Options).
+
+report_sto_results([], _).
+report_sto_results([Type+Result|T], Options) :-
+	print_message(error, plunit(sto(Type, Result))),
+	report_sto_results(T, Options).
+
+
+%%	run_test_6(+Unit, +Name, +Line, +Options, :Body, -Result) is det.
+%
+%	Result is one of:
+%	
+%		* blocked(Unit, Name, Line, Reason)
+%		* failure(Unit, Name, Line, How)
+%		* success(Unit, Name, Line, Determinism, Time)
+%		* setup_failed(Unit, Name, Line)
+
+run_test_6(Unit, Name, Line, Options, _Body,
+	   blocked(Unit, Name, Line, Reason)) :-
+	option(blocked(Reason), Options), !.
+run_test_6(Unit, Name, Line, Options, Body, Result) :-
 	option(all(Answer), Options), !,		% all(Bindings)
-	nondet_test(all(Answer), Unit, Name, Line, Options, Body).
-run_test(Unit, Name, Line, Options, Body) :-
+	nondet_test(all(Answer), Unit, Name, Line, Options, Body, Result).
+run_test_6(Unit, Name, Line, Options, Body, Result) :-
 	option(set(Answer), Options), !,		% set(Bindings)
-	nondet_test(set(Answer), Unit, Name, Line, Options, Body).
-run_test(Unit, Name, Line, Options, Body) :-
+	nondet_test(set(Answer), Unit, Name, Line, Options, Body, Result).
+run_test_6(Unit, Name, Line, Options, Body, Result) :-
 	option(fail, Options), !,			% fail
 	unit_module(Unit, Module),
-	setup(Module, Options), !,
-	statistics(runtime, [T0,_]),
-	(   catch(Module:Body, E, true)
-	->  (   var(E)
-	    ->	statistics(runtime, [T1,_]),
+	(   setup(Module, test(Unit,Name,Line), Options)
+	->  statistics(runtime, [T0,_]),
+	    (   catch(Module:Body, E, true)
+	    ->  (   var(E)
+		->  statistics(runtime, [T1,_]),
+		    Time is (T1 - T0)/1000.0,
+		    Result = failure(Unit, Name, Line, succeeded(Time)),
+		    cleanup(Module, Options)
+		;   Result = failure(Unit, Name, Line, E),
+		    cleanup(Module, Options)
+		)
+	    ;   statistics(runtime, [T1,_]),
 		Time is (T1 - T0)/1000.0,
-		failure(Unit, Name, Line, succeeded(Time), Options),
-		cleanup(Module, Options)
-	    ;	failure(Unit, Name, Line, E, Options),
+		Result = success(Unit, Name, Line, true, Time),
 		cleanup(Module, Options)
 	    )
-	;   statistics(runtime, [T1,_]),
-	    Time is (T1 - T0)/1000.0,
-	    success(Unit, Name, Line, true, Time, Options),
-	    cleanup(Module, Options)
+	;   Result = setup_failed(Unit, Name, Line)
 	).
-run_test(Unit, Name, Line, Options, Body) :-
-	option(true(Cmp), Options),
+run_test_6(Unit, Name, Line, Options, Body, Result) :-
+	option(true(Cmp), Options), !,
 	unit_module(Unit, Module),
-	setup(Module, Options), !,			% true(Binding)
-	statistics(runtime, [T0,_]),
-	(   catch(call_test(Module:Body, Det), E, true)
-	->  (   var(E)
-	    ->	statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-		(   catch(Cmp, _, fail)			% tbd: error
-		->  success(Unit, Name, Line, Det, Time, Options)
-		;   failure(Unit, Name, Line, wrong_answer, Options)
-		),
-		cleanup(Module, Options)
-	    ;	failure(Unit, Name, Line, E, Options),
+	(   setup(Module, test(Unit,Name,Line), Options) % true(Binding)
+	->  statistics(runtime, [T0,_]),
+	    (   catch(call_det(Module:Body, Det), E, true)
+	    ->  (   var(E)
+		->  statistics(runtime, [T1,_]),
+		    Time is (T1 - T0)/1000.0,
+		    (   catch(Cmp, _, fail)			% tbd: error
+		    ->  Result = success(Unit, Name, Line, Det, Time)
+		    ;   Result = failure(Unit, Name, Line, wrong_answer(Cmp))
+		    ),
+		    cleanup(Module, Options)
+		;   Result = failure(Unit, Name, Line, E),
+		    cleanup(Module, Options)
+		)
+	    ;   Result = failure(Unit, Name, Line, failed),
 		cleanup(Module, Options)
 	    )
-	;   failure(Unit, Name, Line, failed, Options),
-	    cleanup(Module, Options)
+	;   Result = setup_failed(Unit, Name, Line)
 	).
-run_test(Unit, Name, Line, Options, Body) :-
-	option(throws(Expect), Options),
+run_test_6(Unit, Name, Line, Options, Body, Result) :-
+	(   option(throws(Expect), Options)
+	->  true
+	;   option(error(ErrorExpect), Options)
+	->  Expect = error(ErrorExpect, _)
+	), !,
 	unit_module(Unit, Module),
-	setup(Module, Options), !,			% true
-	statistics(runtime, [T0,_]),
-	(   catch(Module:Body, E, true)
-	->  (   var(E)
-	    ->	failure(Unit, Name, Line, no_exception, Options),
-		cleanup(Module, Options)
-	    ;	statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-		(   match_error(Expect, E)
-		->  success(Unit, Name, Line, true, Time, Options)
-		;   failure(Unit, Name, Line, wrong_error(Expect, E), Options)
-		),
+	(   setup(Module, test(Unit,Name,Line), Options)
+	->  statistics(runtime, [T0,_]),
+	    (   catch(Module:Body, E, true)
+	    ->  (   var(E)
+		->  Result = failure(Unit, Name, Line, no_exception),
+		    cleanup(Module, Options)
+		;   statistics(runtime, [T1,_]),
+		    Time is (T1 - T0)/1000.0,
+		    (   match_error(Expect, E)
+		    ->  Result = success(Unit, Name, Line, true, Time)
+		    ;   Result = failure(Unit, Name, Line, wrong_error(Expect, E))
+		    ),
+		    cleanup(Module, Options)
+		)
+	    ;   Result = failure(Unit, Name, Line, failed),
 		cleanup(Module, Options)
 	    )
-	;   failure(Unit, Name, Line, failed, Options),
-	    cleanup(Module, Options)
+	;   Result = setup_failed(Unit, Name, Line)
 	).
-run_test(Unit, Name, Line, Options, Body) :-
+run_test_6(Unit, Name, Line, Options, Body, Result) :-
 	unit_module(Unit, Module),
-	setup(Module, Options), !,			% true
-	statistics(runtime, [T0,_]),
-	(   catch(call_test(Module:Body, Det), E, true)
-	->  (   var(E)
-	    ->	statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-		success(Unit, Name, Line, Det, Time, Options),
-		cleanup(Module, Options)
-	    ;	failure(Unit, Name, Line, E, Options),
+	(   setup(Module, test(Unit,Name,Line), Options)
+	->  statistics(runtime, [T0,_]),
+	    (   catch(call_det(Module:Body, Det), E, true)
+	    ->  (   var(E)
+		->  statistics(runtime, [T1,_]),
+		    Time is (T1 - T0)/1000.0,
+		    Result = success(Unit, Name, Line, Det, Time),
+		    cleanup(Module, Options)
+		;   Result = failure(Unit, Name, Line, E),
+		    cleanup(Module, Options)
+		)
+	    ;   Result = failure(Unit, Name, Line, failed),
 		cleanup(Module, Options)
 	    )
-	;   failure(Unit, Name, Line, failed, Options),
-	    cleanup(Module, Options)
+	;   Result = setup_failed(Unit, Name, Line)
 	).
-run_test(_Unit, _Name, _Line, _Options, _Body).
+
 	
-%%	non_det_test(+Expected, +Unit, +Name, +Line, +Options, +Body)
+%%	non_det_test(+Expected, +Unit, +Name, +Line, +Options, +Body, -Result)
 %
 %	Run tests on non-deterministic predicates.
 
-nondet_test(Expected, Unit, Name, Line, Options, Body) :-
+nondet_test(Expected, Unit, Name, Line, Options, Body, Result) :-
 	unit_module(Unit, Module),
-	setup(Module, Options), !,
 	result_vars(Expected, Vars),
 	statistics(runtime, [T0,_]),
-	(   catch(findall(Vars, Module:Body, Bindings), E, true)
-	->  (   var(E)
-	    ->	statistics(runtime, [T1,_]),
-		Time is (T1 - T0)/1000.0,
-	        (   nondet_compare(Expected, Bindings, Unit, Name, Line)
-		->  success(Unit, Name, Line, true, Time, Options)
-		;   failure(Unit, Name, Line, wrong_answer, Options)
-		),
-		cleanup(Module, Options)
-	    ;	failure(Unit, Name, Line, E, Options),
-		cleanup(Module, Options)
+	(   setup(Module, test(Unit,Name,Line), Options)
+	->  (   catch(findall(Vars, Module:Body, Bindings), E, true)
+	    ->  (   var(E)
+		->  statistics(runtime, [T1,_]),
+		    Time is (T1 - T0)/1000.0,
+		    (   nondet_compare(Expected, Bindings, Unit, Name, Line)
+		    ->  Result = success(Unit, Name, Line, true, Time)
+		    ;   Result = failure(Unit, Name, Line, wrong_answer)
+		    ),
+		    cleanup(Module, Options)
+		;   Result = failure(Unit, Name, Line, E),
+		    cleanup(Module, Options)
+		)
 	    )
+	;   Result = setup_failed(Unit, Name, Line)
 	).
 
+
 %%	result_vars(+Expected, -Vars) is det.
 %	
 %	Create a term v(V1, ...) containing all variables at the left
@@ -570,81 +829,86 @@ cmp_list([E0|ET], [V0|VT], Op) :-
 
 cmp(Var  == Value, Var,  ==, Value).
 cmp(Var =:= Value, Var, =:=, Value).
-cmp(Var =@= Value, Var, =@=, Value).
 cmp(Var  =  Value, Var,  =,  Value).
+:- if(swi).
+cmp(Var =@= Value, Var, =@=, Value).
+:- else.
+:- if(sicstus).
+cmp(Var =@= Value, Var, variant, Value). % variant/2 is the same =@=
+:- endif.
+:- endif.
 
 
-%%	call_test(:Goal, -Det) is nondet.
+%%	call_det(:Goal, -Det) is nondet.
 %
 %	True if Goal succeeded.  Det is unified to =true= if Goal left
 %	no choicepoints and =false= otherwise.
 
-:- if(swi).
-call_test(Goal, Det) :-
-	Goal,
-	deterministic(Det).
+:- if((swi|sicstus)).
+call_det(Goal, Det) :-
+	call_cleanup(Goal,Det0=true),
+	( var(Det0) -> Det = false ; Det = true ).
 :- else.
-:- if(sicstus).
-call_test(Goal, Det) :-
-	statistics(choice, [Used0|_]),
-	Goal,
-	statistics(choice, [Used1|_]),
-	(   Used1 =:= Used0
-	->  Det = true
-	;   Det = false
-	).
-:- else.
-call_test(Goal, true) :-
+call_det(Goal, true) :-
 	call(Goal).
 :- endif.
-:- endif.
 
 %%	match_error(+Expected, +Received) is semidet.
+%
+%	True if the Received errors matches the expected error. Matching
+%	is based on subsumes_chk/2.
 
 match_error(Expect, Rec) :-
-	Expect =@= Rec, !.
-match_error(error(FormalEx, ContextEx), error(FormalRec, ContextRec)) :-
-	FormalEx =@= FormalRec,
-	ContextEx = ContextRec.
+	subsumes_chk(Expect, Rec).
 
-%%	setup(+Module, +Options) is semidet.
+%%	setup(+Module, +Context, +Options) is semidet.
 %
 %	Call the setup handler and  fail  if   it  cannot  run  for some
 %	reason. The condition handler is  similar,   but  failing is not
-%	considered an error.
+%	considered an error.  Context is one of
+%	
+%	    * unit(Unit)
+%	    If it is the setup handler for a unit
+%	    * test(Unit,Name,Line)
+%	    If it is the setup handler for a test
 
-setup(Module, Options) :-
+setup(Module, Context, Options) :-
 	option(setup(Setup), Options), !,
 	(   catch(Module:Setup, E, true)
 	->  (   var(E)
 	    ->	true
-	    ;	print_message(error, E), 	% TBD
+	    ;	print_message(error, plunit(error(setup, Context, E))),
 		fail
 	    )
-	;   print_message(error, goal_failed(Setup)),
+	;   print_message(error, error(goal_failed(Setup), _)),
 	    fail
 	).
-setup(Module, Options) :-
+setup(Module, Context, Options) :-
 	option(condition(Setup), Options), !,
 	(   catch(Module:Setup, E, true)
 	->  (   var(E)
 	    ->	true
-	    ;	print_message(error, E), 	% TBD
+	    ;	print_message(error, plunit(error(condition, Context, E))),
 		fail
 	    )
 	;   fail
 	).
-setup(_,_).
+setup(_,_,_).
 
 %%	cleanup(+Module, +Options) is det.
 %
-%	Call the cleanup handler and succeed,   regardless of any errors
-%	or failure without notice.
+%	Call the cleanup handler and succeed.   Failure  or error of the
+%	cleanup handler is reported, but tests continue normally.
 
 cleanup(Module, Options) :-
 	option(cleanup(Cleanup), Options, true),
-	ignore(catch(Module:Cleanup, _, true)).
-
+	(   catch(Module:Cleanup, E, true)
+	->  (   var(E)
+	    ->	true
+	    ;	print_message(warning, E)
+	    )
+	;   print_message(warning, goal_failed(Cleanup, '(cleanup handler)'))
+	).
 
 success(Unit, Name, Line, Det, Time, Options) :-
 	assert(passed(Unit, Name, Line, Det, Time)),
@@ -658,8 +922,34 @@ success(Unit, Name, Line, Det, Time, Options) :-
 	flush_output(user_error).
 
 failure(Unit, Name, Line, E, _Options) :-
-	assert(failed(Unit, Name, Line, E)),
-	report_failure(Unit, Name, Line, E).
+	report_failure(Unit, Name, Line, E),
+	assert_cyclic(failed(Unit, Name, Line, E)).
+
+%%	assert_cyclic(+Term) is det.
+%
+%	Assert  a  possibly  cyclic  unit   clause.  Current  SWI-Prolog
+%	assert/1 does not handle cyclic terms,  so we emulate this using
+%	the recorded database.
+%	
+%	@tbd	Implement cycle-safe assert and remove this.
+
+:- if(swi).
+assert_cyclic(Term) :-
+	acyclic_term(Term), !,
+	assert(Term).
+assert_cyclic(Term) :-
+	Term =.. [Functor|Args],
+	recorda(cyclic, Args, Id),
+	functor(Term, _, Arity),
+	length(NewArgs, Arity),
+	Head =.. [Functor|NewArgs],
+	assert((Head :- recorded(_, Var, Id), Var = NewArgs)).
+:- else.
+:- if(sicstus).
+:- endif.
+assert_cyclic(Term) :-
+	assert(Term).
+:- endif.
 
 
 		 /*******************************
@@ -672,12 +962,30 @@ failure(Unit, Name, Line, E, _Options) :-
 %	them to current output and fail.
 
 report :-
-	report_blocked,
-	report_failed.
+	number_of_clauses(passed/5, Passed),
+	number_of_clauses(failed/4, Failed),
+	number_of_clauses(blocked/4, Blocked),
+	number_of_clauses(sto/4, STO),
+	(   Passed+Failed+Blocked+STO =:= 0
+	->  info(plunit(no_tests))
+	;   Failed+Blocked+STO =:= 0
+	->  info(plunit(all_passed(Passed)))
+	;   report_blocked,
+	    report_failed,
+	    report_sto
+	).
+
+number_of_clauses(F/A,N) :-
+	(	current_predicate(F/A)
+	->	functor(G,F,A),
+		findall(t, G, Ts),
+		length(Ts, N)
+	;	N = 0
+	).
 
 report_blocked :-
-	predicate_property(blocked(_,_,_,_), number_of_clauses(N)),
-	N > 0,
+	number_of_clauses(blocked/4,N),
+	N > 0, !,
 	info(plunit(blocked(N))),
 	(   blocked(Unit, Name, Line, Reason),
 	    unit_file(Unit, File),
@@ -688,13 +996,21 @@ report_blocked :-
 report_blocked.
 
 report_failed :-
-	predicate_property(failed(_,_,_,_), number_of_clauses(N)),
+	number_of_clauses(failed/4,N),
 	N > 0, !,
 	info(plunit(failed(N))),
 	fail.
 report_failed :-
 	info(plunit(failed(0))).
 
+report_sto :-
+	number_of_clauses(sto/4,N),
+	N > 0, !,
+	info(plunit(sto(N))),
+	fail.
+report_sto :-
+	info(plunit(sto(0))).
+
 report_failure(Unit, Name, Line, Error) :-
 	print_message(error, plunit(failed(Unit, Name, Line, Error))).
 
@@ -771,14 +1087,23 @@ message_level(Level) :-
 	;   Level = silent
 	).
 
+locationprefix(File:Line) --> !,
+	[ '~w:~d:\n\t'-[File,Line]].
+locationprefix(test(Unit,_Test,Line)) --> !,
+	{ unit_file(Unit, File) },
+	locationprefix(File:Line).
+locationprefix(unit(Unit)) --> !,
+	[ 'PL-Unit: unit ~w: '-[Unit] ].
+locationprefix(FileLine) -->
+	{ throw_error(type_error(locationprefix,FileLine), _) }.
 
 message(error(context_error(plunit_close(Name, -)), _)) -->
 	[ 'PL-Unit: cannot close unit ~w: no open unit'-[Name] ].
 message(error(context_error(plunit_close(Name, Start)), _)) -->
 	[ 'PL-Unit: cannot close unit ~w: current unit is ~w'-[Name, Start] ].
 message(plunit(nondet(File, Line, Name))) -->
-	[ '~w:~d: PL-Unit: Test ~w: Test succeeded with choicepoint'-
-	  [File, Line, Name] ].
+	locationprefix(File:Line),
+	[ 'PL-Unit: Test ~w: Test succeeded with choicepoint'- [Name] ].
 					% Unit start/end
 :- if(swi).
 message(plunit(begin(Unit))) -->
@@ -795,23 +1120,119 @@ message(plunit(blocked(unit(Unit, Reason)))) -->
 	[ 'PL-Unit: ~w blocked: ~w'-[Unit, Reason] ].
 
 					% Blocked tests
+message(plunit(blocked(1))) --> !,
+	[ 'one test is blocked:'-[] ].
 message(plunit(blocked(N))) -->
-	[ '~D tests where blocked'-[N] ].
+	[ '~D tests are blocked:'-[N] ].
 message(plunit(blocked(Pos, Name, Reason))) -->
-	[ '  ~w: test ~w: ~w'-[Pos, Name, Reason] ].
+	locationprefix(Pos),
+	[ 'test ~w: ~w'-[Name, Reason] ].
 
 					% fail/success
+message(plunit(no_tests)) --> !,
+	[ 'No tests to run' ].
+message(plunit(all_passed(Count))) --> !,
+	[ 'All ~D tests passed'-[Count] ].
 message(plunit(failed(0))) --> !,
-	[ 'All tests passed'-[] ].
+	[].
+message(plunit(failed(1))) --> !,
+	[ '1 test failed'-[] ].
 message(plunit(failed(N))) -->
 	[ '~D tests failed'-[N] ].
-message(plunit(failed(Unit, Name, Line, succeeded(Time)))) -->
+message(plunit(sto(0))) --> !,
+	[].
+message(plunit(sto(N))) -->
+	[ '~D test results depend on unification mode'-[N] ].
+message(plunit(failed(Unit, Name, Line, Failure))) -->
        { unit_file(Unit, File) },
-       [ '~w:~w: test ~w: must fail but succeeded in ~2f seconds~n'-
-	 [File, Line, Name, Time] ].
-message(plunit(failed(Unit, Name, Line, Error))) -->
+       locationprefix(File:Line),
+       ['test ~w: '- [Name] ],
+       failure(Failure).
+					% Setup/condition errors
+message(plunit(error(Where, Context, Exception))) -->
+	locationprefix(Context),
+	{ message_to_string(Exception, String) },
+	[ 'error in ~w: ~w'-[Where, String] ].
+
+					% STO messages
+message(plunit(sto(Unit, Name, Line))) -->
 	{ unit_file(Unit, File) },
-	[ '~w:~w: test ~w: ~p~n'-[File, Line, Name, Error] ].
+       locationprefix(File:Line),
+       ['test ~w is subject to occurs check (STO): '- [Name] ].
+message(plunit(sto(Type, Result))) -->
+	sto_type(Type),
+	sto_result(Result).
+
+sto_type(sto_error_incomplete) -->
+	[ 'Finite trees (error checking): ' ].
+sto_type(rational_trees) -->
+	[ 'Rational trees: ' ].
+sto_type(finite_trees) -->
+	[ 'Finite trees: ' ].
+
+sto_result(success(_Unit, _Name, _Line, Det, Time)) -->
+	det(Det),
+	[ ' success in ~2f seconds'-[Time] ].
+sto_result(failure(_Unit, _Name, _Line, How)) -->
+	failure(How).
+
+det(true) -->
+	[ 'deterministic' ].
+det(false) -->
+	[ 'non-deterministic' ].
+
+:- if(swi).
+write_term(T, OPS) -->
+	['~@'-[write_term(T,OPS)]].
+:- else.
+write_term(T, _OPS) -->
+	['~q'-[T]].
+:- endif.
+
+expected_got_ops_(Ex, E, OPS, Goals) -->
+	['    Expected: '-[]], write_term(Ex, OPS), [nl],
+	['    Got:      '-[]], write_term(E,  OPS), [nl],
+	( { Goals = [] } -> []
+	; ['       with: '-[]], write_term(Goals, OPS), [nl]
+	).
+
+
+failure(Var) -->
+	{ var(Var) }, !,
+	[ 'Unknown failure?' ].
+failure(succeeded(Time)) --> !,
+	[ 'must fail but succeeded in ~2f seconds~n'-[Time] ].
+failure(wrong_error(Expected, Error)) --> !,
+	{ copy_term(Expected-Error, Ex-E, Goals),
+	  numbervars(Ex-E-Goals, 0, _),
+	  write_options(OPS)
+	},
+	[ 'wrong error'-[], nl ],
+	expected_got_ops_(Ex, E, OPS, Goals).
+failure(wrong_answer(Cmp)) -->
+	{ Cmp =.. [Op,Answer,Expected], !,
+	  copy_term(Expected-Answer, Ex-A, Goals),
+	  numbervars(Ex-A-Goals, 0, _),
+	  write_options(OPS)
+	},
+	[ 'wrong answer (compared using ~w)'-[Op], nl ],
+	expected_got_ops_(Ex, A, OPS, Goals).
+:- if(swi).
+failure(Error) -->
+	{ Error = error(_,_), !,
+	  message_to_string(Error, Message)
+	},
+	[ 'received error: ~w'-[Message] ].
+:- endif.
+failure(Why) -->
+	[ '~p~n'-[Why] ].
+
+write_options([ numbervars(true),
+		quoted(true),
+		portray(true),
+		max_depth(10),
+		attributes(portray)
+	      ]).
 
 :- if(swi).
 
@@ -836,7 +1257,7 @@ user:generate_message_hook(Message) -->
 	message(Message),
 	[nl].				% SICStus requires nl at the end
 
-%	user:message_hook(+Sevirity, +Message, +Lines) is semidet.
+%	user:message_hook(+Severity, +Message, +Lines) is semidet.
 %
 %	Redefine printing some messages. It appears   SICStus has no way
 %	to get multiple messages at the same   line, so we roll our own.
