@@ -1,4 +1,4 @@
-/*  $Id: rdf_db.pl,v 1.107 2007/07/09 13:56:38 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -50,6 +50,7 @@
 	    rdf_set_predicate/2,	% +Predicate, +Property
 	    rdf_predicate_property/2,	% +Predicate, ?Property
 	    rdf_current_predicate/1,	% -Predicate
+	    rdf_current_literal/1,	% -Literal
 	    rdf_transaction/1,		% :Goal
 	    rdf_transaction/2,		% :Goal, +Id
 	    rdf_active_transaction/1,	% ?Id
@@ -107,6 +108,7 @@
 	    rdf_destroy_literal_map/1,	% +Handle
 	    rdf_reset_literal_map/1,	% +Handle
 	    rdf_insert_literal_map/3,	% +Handle, +Key, +Literal
+	    rdf_insert_literal_map/4,	% +Handle, +Key, +Literal, -NewKeys
 	    rdf_delete_literal_map/3,	% +Handle, +Key, +Literal
 	    rdf_delete_literal_map/2,	% +Handle, +Key
 	    rdf_find_literal_map/3,	% +Handle, +KeyList, -Literals
@@ -126,6 +128,7 @@
 :- use_module(library(nb_set)).
 :- use_module(library(error)).
 :- use_module(library(url)).
+:- use_module(library(debug)).
 :- use_module(rdf_cache).
 
 :- initialization
@@ -197,6 +200,31 @@ rdf_register_ns(Alias, URI, Options) :-
 rdf_register_ns(Alias, URI, _) :-
 	assert(ns(Alias, URI)).
 
+%%	register_file_ns(+Map:list(NS=URL)) is det.
+%	
+%	Register a namespace as encounted in   the  namespace list of an
+%	RDF document. We only register if  both the abbreviation and URL
+%	are not already known. Is there a   better  way? This code could
+%	also do checks on the consistency   of  RDF and other well-known
+%	namespaces.
+%	
+%	@tbd	Better error handling
+
+register_file_ns([]) :- !.
+register_file_ns([NS=URL|T]) :- !,
+	register_file_ns(NS=URL),
+	register_file_ns(T).
+register_file_ns([]=_) :- !.		% xmlns= (overall default)
+register_file_ns(NS=URL) :-
+	(   rdf_db:ns(NS, URL)
+	->  true
+	;   rdf_db:ns(NS, _)
+	->  true			% redefined abbreviation
+	;   rdf_db:ns(_, URL)
+	->  true			% redefined URL
+	;   rdf_register_ns(NS, URL)
+	).
+
 
 %%	rdf_global_id(?Id, ?GlobalId) is det.
 %
@@ -208,11 +236,14 @@ rdf_global_id(NS:Local, Global) :-
 rdf_global_id(Global, Global).
 
 
-%%	rdf_global_object(?Object, ?GlobalObject) is det.
+%%	rdf_global_object(+Object, -GlobalObject) is semidet.
+%%	rdf_global_object(-Object, +GlobalObject) is semidet.
 %	
 %	Same as rdf_global_id/2,  but  intended   for  dealing  with the
 %	object part of a  triple,  in   particular  the  type  for typed
 %	literals.
+%	
+%	@error	existence_error(rdf_namespace, NS)
 
 rdf_global_object(NS:Local, Global) :-
 	global(NS, Local, Global), !.
@@ -227,13 +258,12 @@ global(NS, Local, Global) :-
 	    atom_concat(Full, Local, Global)
 	;   atom(NS), atom(Local)
 	->  (   ns(NS, Full)
-	    *-> atom_concat(Full, Local, Global)
-	    ;   atom_concat(NS, Local, Global)
+	    *->	atom_concat(Full, Local, Global)
+	    ;	existence_error(rdf_namespace, NS)
 	    )
 	).
 
 
-
 %%	rdf_global_term(+TermIn, -GlobalTerm) is det.
 %	
 %	Does rdf_global_id/2 on all terms NS:Local by recursively analysing
@@ -242,8 +272,8 @@ global(NS, Local, Global) :-
 rdf_global_term(Var, Var) :-
 	var(Var), !.
 rdf_global_term(NS:Local, Global) :-
-	rdf_global_id(NS:Local, Global0), !,
-	Global = Global0.
+	atom(NS), atom(Local), ns(NS, Full), !,
+	atom_concat(Full, Local, Global).
 rdf_global_term([H0|T0], [H|T]) :- !,
 	rdf_global_term(H0, H),
 	rdf_global_term(T0, T).
@@ -428,7 +458,7 @@ rdf_member_property(P, N) :-
 %
 %	Generate a unique blank node identifier for a subject.
 %	
-%	@depricated	New code should use rdf_bnode/1.
+%	@deprecated	New code should use rdf_bnode/1.
 
 rdf_node(Resource) :-
 	rdf_bnode(Resource).
@@ -657,17 +687,28 @@ monitor_mask(all,	   0xffff).
 
 rdf_save_db(File) :-
 	open(File, write, Out, [type(binary)]),
+	set_stream(Out, record_position(false)),
 	call_cleanup(rdf_save_db_(Out, _), close(Out)).
 
 
 rdf_save_db(File, DB) :-
 	open(File, write, Out, [type(binary)]),
+	set_stream(Out, record_position(false)),
 	call_cleanup(rdf_save_db_(Out, DB), close(Out)).
 
 
-rdf_load_db_no_admin(File, Id) :-
-	open(File, read, Out, [type(binary)]),
-	call_cleanup(rdf_load_db_(Out, Id), close(Out)).
+%%	rdf_load_db_no_admin(+File, +Id, -Graphs) is det.
+%
+%	Load triples from a  .trp  file   without  updating  the  source
+%	administration. Id is  handled  to   monitor  action.  Graphs is
+%	either an atom, indicating a single loaded   graph  or a list of
+%	graph-names encountered in File.
+
+rdf_load_db_no_admin(File, Id, Graphs) :-
+	open(File, read, In, [type(binary)]),
+	set_stream(In, record_position(false)),
+	call_cleanup(rdf_load_db_(In, Id, Graphs), close(In)).
+
 
 %%	rdf_load_db(+File) is det.
 %
@@ -676,9 +717,11 @@ rdf_load_db_no_admin(File, Id) :-
 
 rdf_load_db(File) :-
 	file_name_to_url(File, URL),
-	rdf_load_db_no_admin(File, URL),
-	rdf_graphs_(DBList),
-	(   member(DB, DBList),
+	rdf_load_db_no_admin(File, URL, Graphs),
+	(   (   is_list(Graphs)
+	    ->	member(DB, Graphs)
+	    ;	DB = Graphs
+	    ),
 	    rdf_md5(DB, MD5),
 	    rdf_statistics_(triples(DB, Triples)),
 	    retractall(rdf_source(DB, _, _, _, _)),
@@ -695,9 +738,10 @@ rdf_load_db(File) :-
 :- multifile
 	rdf_open_hook/3,
 	rdf_load_stream/3,
-	rdf_input_info/3,
+	rdf_input_info_hook/3,
 	rdf_file_type/2,
-	url_protocol/1.
+	url_protocol/1,
+	exists_url/1.
 
 %%	rdf_load(+FileOrList) is det.
 %%	rdf_load(+FileOrList, +Options) is det.
@@ -730,6 +774,10 @@ rdf_load_db(File) :-
 %	    * cache(Bool)
 %	    If =false=, do not use or create a cache file.
 %	    
+%	    * register_namespaces(Bool)
+%	    If =true= (default =false=), register xmlns= namespace
+%	    declarations as ns/2 namespaces if there is no conflict.
+%	    
 %	Other options are forwarded to process_rdf/3.
 
 rdf_load(Spec) :-
@@ -739,7 +787,8 @@ rdf_load([], _) :- !.
 rdf_load([H|T], Options) :- !,
 	rdf_load(H, Options),
 	rdf_load(T, Options).
-rdf_load(Spec, Options) :-
+rdf_load(Spec, Options0) :-
+	fix_options(Options0, Options),
 	statistics(cputime, T0),
 	(   select(result(Action, Triples, MD5), Options, Options1)
 	->  true
@@ -756,27 +805,34 @@ rdf_load(Spec, Options) :-
 	select_option(blank_nodes(ShareMode), Options3, Options4, share),
 	select_option(cache(Cache), Options4, Options5, true),
 	select_option(if(If), Options5, Options6, changed),
-	select_option(db(DB), Options6, RDFOptions, SourceURL),
+	select_option(db(DB), Options6, Options7, SourceURL),
+	select_option(register_namespaces(RegNS), Options7, RDFOptions0, false),
 	(   var(BaseURI)
 	->  BaseURI = SourceURL
 	;   true
 	),
+	(   RegNS == true
+	->  RDFOptions = [ namespaces(NSList)|RDFOptions0]
+	;   RDFOptions = RDFOptions0
+	),
 	(   must_load(If, DB, Modified)
 	->  do_unload(DB),		% unload old
 	    (   Cache == true,
 		read_cache(SourceURL, Modified, CacheFile),
-	        catch(rdf_load_db_no_admin(CacheFile, cache(DB)), _, fail)
+	        catch(rdf_load_db_no_admin(CacheFile, cache(DB), _Graphs), _, fail)
 	    ->	Action = load
 	    ;   rdf_input_open(Input, Stream, Format),
 		must_be(ground, Format),
-		call_cleanup(rdf_load_stream(Format, Stream,
+		rdf_format(Format, RDFFormat),
+		rdf_set_graph_source(DB, SourceURL),
+		call_cleanup(rdf_load_stream(RDFFormat, Stream,
 					     [ base_uri(BaseURI),
 					       blank_nodes(ShareMode),
 					       db(DB)
 					     | RDFOptions
 					     ]),
 			     close_input(Input, Stream)), !,
-		rdf_set_graph_source(DB, SourceURL),
+		register_file_ns(NSList),
 		(   Cache == true,
 		    rdf_cache_file(SourceURL, write, CacheFile)
 		->  catch(save_cache(DB, CacheFile), E,
@@ -796,6 +852,25 @@ rdf_load(Spec, Options) :-
 	),
 	report_loaded(Action, Input, DB, Triples, T0, Options).
 
+%%	fix_options(+Spec, -Options) is det.
+%
+%	Fix options for old format
+
+fix_options(DB, Options) :-
+	atom(DB), DB \== [], !,
+	Options = [db(DB)].
+fix_options(Options, Options).
+
+%%	rdf_format(+FormatIn, -Format)
+%
+%	Remove format envelopes, such as gzip(Format).
+
+rdf_format(Format, Format) :-
+	atom(Format), !.
+rdf_format(Format0, Format) :-
+	arg(1, Format0, Format1),
+	rdf_format(Format1, Format).
+
 %%	close_input(+Input, +Stream) is det.
 %
 %	Close input if it was not specified as a stream.
@@ -804,7 +879,7 @@ close_input(stream(_), _) :- !.
 close_input(_, Stream) :-
 	close(Stream).
 
-%%	rdf_input(+Term, -Input, -BaseURI)
+%%	rdf_input(+Term, -Input, -BaseURI) is semidet.
 %
 %	Resolve input description Term.  Unify   Input  with a canonical
 %	description of the input, which is one of:
@@ -815,6 +890,9 @@ close_input(_, Stream) :-
 %	
 %	BaseURI is unified with a default base   URI. Note that this may
 %	be overruled from the options of rdf_load/2.
+%	
+%	@tbd	More generic mechanism to deal with encoding envelopes
+%		such as gzip compression.
 
 rdf_input(stream(Stream), stream(Stream), BaseURI) :- !,
 	(   stream_property(Stream, file_name(File))
@@ -824,18 +902,42 @@ rdf_input(stream(Stream), stream(Stream), BaseURI) :- !,
 rdf_input(Stream, stream(Stream), BaseURI) :-
 	is_stream(Stream), !,
 	rdf_input(stream(Stream), _, BaseURI).
-rdf_input(FileURL, file(File), FileURL) :-
+rdf_input(FileURL, file(File), BaseURI) :-
 	atom(FileURL),
-	file_name_to_url(File, FileURL), !.
-rdf_input(URL, url(Protocol, URL), URL) :-
-	is_url(URL, Protocol), !.
+	file_name_to_url(File0, FileURL), !,
+	file_input(File0, File, BaseURI).
+rdf_input(URL0, url(Protocol, URL), URL0) :-
+	is_url(URL0, Protocol), !,
+	(   gzip_extend(URL0, URL),
+	    exists_url(url(Protocol, URL))
+	->  true
+	;   URL=URL0
+	).
 rdf_input(Spec, file(Path), BaseURI) :-
-	findall(Ext, (rdf_file_type(Ext, _);Ext=''), Exts),
+	file_input(Spec, Path, BaseURI).
+
+file_input(Spec, Path, BaseURI) :-
+	findall(Ext, (rdf_file_type(Ext, _);Ext=gz;Ext=''), Exts),
 	absolute_file_name(Spec, Path,
 			   [ access(read),
-			     extensions(Exts)
+			     extensions(Exts),
+			     file_errors(fail)
 			   ]),
-	file_name_to_url(Path, BaseURI).
+	file_name_to_url(Path, BaseURI0),
+	clean_base_uri(BaseURI0, BaseURI).
+
+gzip_extend(Path, Path).
+gzip_extend(Path0, Path) :-
+	file_name_extension(Path0, gz, Path).
+
+%%	clean_base_uri(+BaseURI0, -BaseURI) is det.
+%
+%	BaseURI  is  BaseURI0  after    removing  packaging  extensions.
+%	Currently only deals with =|.gz|=.
+
+clean_base_uri(BaseURI0, BaseURI) :-
+	file_name_extension(BaseURI, gz, BaseURI0), !.
+clean_base_uri(BaseURI, BaseURI).
 	
 %%	is_url(+Term, -Protocol) is semidet.
 %
@@ -850,6 +952,12 @@ is_url(URL, Protocol) :-
 	downcase_atom(RawProtocol, Protocol),
 	url_protocol(Protocol).
 
+%%	exists_url(+URL) is semidet.
+%
+%	True if URL exists.
+
+%	(no default clauses)
+
 %%	url_protocol(+Protocol) is det.
 %
 %	True  if  Protocol  is  the  (lowercase)  name  of  a  supported
@@ -864,6 +972,8 @@ is_url(URL, Protocol) :-
 %	Return the last modification  time  of   Input  as  a POSIX time
 %	stamp as well as the format of the input.
 
+rdf_input_info(Input, Modified, Format) :-
+	rdf_input_info_hook(Input, Modified, Format), !.
 rdf_input_info(file(File), Modified, Format) :-
 	time_file(File, Modified),
 	file_name_extension(_, Ext, File),
@@ -914,7 +1024,7 @@ rdf_load_stream(xhtml, Stream, Options) :- !,
 			parse(Id)).
 rdf_load_stream(triples, Stream, Options) :- !,
 	option(db(Id), Options),
-	rdf_load_db_(Stream, Id).
+	rdf_load_db_(Stream, Id, _Graphs).
 
 %%	read_cache(+BaseURI, +SourceModified, -CacheFile) is semidet.
 %
@@ -1007,7 +1117,7 @@ rdf_source(DB, SourceURL) :-
 %	
 %	True if Source is a loaded source.
 %	
-%	@depricated	Use rdf_graph/1 or rdf_source/2.
+%	@deprecated	Use rdf_graph/1 or rdf_source/2.
 
 rdf_source(DB) :-
 	rdf_source(DB,_,_,_,_).		% loaded files
@@ -1464,7 +1574,8 @@ rdf_save_subject(Out, Subject, BaseURI, Atts, Indent, Options) :-
 	select(RdfType=Type, Atts, Atts1),
 	rdf_id(Type, BaseURI, TypeId),
 	xml_is_name(TypeId), !,
-	format(Out, '~*|<~w', [Indent, TypeId]),
+	format(Out, '~*|<', [Indent]),
+	rdf_write_id(Out, TypeId),
 	save_about(Out, BaseURI, Subject),
 	save_attributes(Atts1, BaseURI, Out, TypeId, Indent, Options).
 rdf_save_subject(Out, Subject, BaseURI, Atts, Indent, Options) :-
@@ -1507,7 +1618,9 @@ save_attributes(Atts, BaseURI, Out, Element, Indent, Options) :-
 	->  format(Out, '/>~n', [])
 	;   format(Out, '>~n', []),
 	    save_attributes2(InBody, BaseURI, body, Out, SubIndent, Options),
-	    format(Out, '~N~*|</~w>~n', [Indent, Element])
+	    format(Out, '~N~*|</', [Indent]),
+	    rdf_write_id(Out, Element),
+	    format(Out, '>~n', [])
 	).
 
 %%	split_attributes(+Attributes, -HeadAttrs, -BodyAttr)
@@ -1576,7 +1689,9 @@ save_attribute(tag, Name=literal(Value), BaseURI, Out, Indent, _DB) :-
 	rdf_id(Name, BaseURI, NameText),
 	stream_property(Out, encoding(Encoding)),
 	xml_quote_attribute(Value, QVal, Encoding),
-	format(Out, '~N~*|~w="~w"', [AttIndent, NameText, QVal]).
+	format(Out, '~N~*|', [AttIndent]),
+	rdf_write_id(Out, NameText),
+	format(Out, '="~w"', [QVal]).
 save_attribute(body, Name=literal(Literal0), BaseURI, Out, Indent, Options) :- !,
 	rdf_id(Name, BaseURI, NameText),
 	(   memberchk(convert_typed_literal(Converter), Options),
@@ -1584,37 +1699,36 @@ save_attribute(body, Name=literal(Literal0), BaseURI, Out, Indent, Options) :- !
 	->  Literal = type(Type, Content)
 	;   Literal = Literal0
 	),
+	format(Out, '~N~*|<', [Indent]),
+	rdf_write_id(Out, NameText),
 	(   Literal = lang(Lang, Value)
 	->  (   memberchk(document_language(Lang), Options)
-	    ->  format(Out, '~N~*|<~w>', [Indent, NameText])
+	    ->  write(Out, '>')
 	    ;   rdf_id(Lang, BaseURI, LangText),
-		format(Out, '~N~*|<~w xml:lang="~w">',
-		       [Indent, NameText, LangText])
+		format(Out, ' xml:lang="~w">', [LangText])
 	    )
 	;   Literal = type(Type, Value)
 	->  (   rdf_equal(Type, rdf:'XMLLiteral')
-	    ->	format(Out, '~N~*|<~w rdf:parseType="Literal">',
-		       [Indent, NameText])
+	    ->	write(Out, ' rdf:parseType="Literal">')
 	    ;	stream_property(Out, encoding(Encoding)),
 		rdf_value(Type, BaseURI, QVal, Encoding),
-		format(Out, '~N~*|<~w rdf:datatype="~w">',
-		       [Indent, NameText, QVal])
+		format(Out, ' rdf:datatype="~w">', [QVal])
 	    )
 	;   atomic(Literal)
-	->  format(Out, '~N~*|<~w>', [Indent, NameText]),
+	->  write(Out, '>'),
 	    Value = Literal
-	;   format(Out, '~N~*|<~w rdf:parseType="Literal">',
-		   [Indent, NameText]),
+	;   write(Out, ' rdf:parseType="Literal">'),
 	    Value = Literal
 	),
 	save_attribute_value(Value, Out, Indent),
-	format(Out, '</~w>', [NameText]).
+	write(Out, '</'), rdf_write_id(Out, NameText), write(Out, '>').
 save_attribute(body, Name=Value, BaseURI, Out, Indent, Options) :-
 	rdf_is_bnode(Value), !,
 	rdf_id(Name, BaseURI, NameText),
+	format(Out, '~N~*|<', [Indent]),
+	rdf_write_id(Out, NameText),
 	(   named_anon(Value, NodeID)
-	->  format(Out, '~N~*|<~w rdf:nodeID="~w"/>',
-		   [Indent, NameText, NodeID])
+	->  format(Out, ' rdf:nodeID="~w"/>', [NodeID])
 	;   (   rdf(S1, Name, Value),
 	        rdf(S2, P2, Value),
 		(S1 \== S2 ; Name \== P2)
@@ -1625,20 +1739,23 @@ save_attribute(body, Name=Value, BaseURI, Out, Indent, Options) :-
 	    ),
 	    SubIndent is Indent + 2,
 	    (   rdf(Value, rdf:type, rdf:'List')
-	    ->  format(Out, '~N~*|<~w', [Indent, NameText]),
-		save_about(Out, BaseURI, Value),
+	    ->  save_about(Out, BaseURI, Value),
 		format(Out, ' rdf:parseType="Collection">~n', []),
 		rdf_save_list(Out, Value, BaseURI, SubIndent, Options)
-	    ;   format(Out, '~N~*|<~w>~n', [Indent, NameText]),
+	    ;   format(Out, '>~n', []),
 		rdf_save_subject(Out, Value, BaseURI, SubIndent, Options)
 	    ),
-	    format(Out, '~N~*|</~w>~n', [Indent, NameText])
+	    format(Out, '~N~*|</', [Indent]),
+	    rdf_write_id(Out, NameText),
+	    format(Out, '>~n', [])
 	).
 save_attribute(body, Name=Value, BaseURI, Out, Indent, _DB) :-
 	stream_property(Out, encoding(Encoding)),
 	rdf_value(Value, BaseURI, QVal, Encoding),
 	rdf_id(Name, BaseURI, NameText),
-	format(Out, '~N~*|<~w rdf:resource="~w"/>', [Indent, NameText, QVal]).
+	format(Out, '~N~*|<', [Indent]),
+	rdf_write_id(Out, NameText),
+	format(Out, ' rdf:resource="~w"/>', [QVal]).
 
 save_attribute_value(Value, Out, _) :-	% strings
 	atom(Value), !,
@@ -1697,6 +1814,16 @@ rdf_id(Id, _, NS:Local) :-
 rdf_id(Id, _, Id).
 
 
+%%	rdf_write_id(+Out, +NSLocal) is det.
+%
+%	Write an identifier. We cannot use native write on it as both NS
+%	and Local can be operators.
+
+rdf_write_id(Out, NS:Local) :- !,
+	format(Out, '~w:~w', [NS, Local]).
+rdf_write_id(Out, Atom) :-
+	write(Out, Atom).
+
 %%	rdf_value(+Resource, +BaseURI, -Text, +Encoding)
 %	
 %	According  to  "6.4  RDF  URI  References"  of  the  RDF  Syntax
