@@ -1,4 +1,4 @@
-/*  $Id: pl-funcs.h,v 1.289 2007/06/26 08:55:07 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -62,7 +62,6 @@ COMMON(word) 		consPtr(void *p, word ts);
 #endif
 COMMON(char *) 		store_string(const char *s);
 COMMON(void) 		remove_string(char *s);
-COMMON(int) 		unboundStringHashValue(const char *t, size_t l);
 #ifndef xmalloc
 COMMON(void *) 		xmalloc(size_t size);
 COMMON(void *) 		xrealloc(void *mem, size_t size);
@@ -81,11 +80,10 @@ COMMON(void)		freezeGlobal(ARG1_LD);
 
 /* pl-wam.c */
 COMMON(word) 		pl_count(void);
-COMMON(void) 		TrailAssignment(Word p);
+COMMON(void) 		TrailAssignment__LD(Word p ARG_LD);
 COMMON(void) 		do_undo(mark *m);
 COMMON(Definition) 	getProcDefinition(Procedure proc);
 COMMON(void) 		fix_term_ref_count(void);
-COMMON(word) 		pl_unify_with_occurs_check(term_t t1, term_t t2);
 COMMON(fid_t) 		PL_open_signal_foreign_frame(void);
 COMMON(bool)		foreignWakeup(ARG1_LD);
 
@@ -130,10 +128,10 @@ COMMON(void)		pushArithStack(Number n ARG_LD);
 COMMON(void)		resetArithStack(ARG1_LD);
 COMMON(Number)		argvArithStack(int n ARG_LD);
 COMMON(void)		popArgvArithStack(int n ARG_LD);
+COMMON(int)		ar_sign_i(Number n1);
 
 /* pl-bag.c */
-COMMON(void) 		resetBags(void);
-COMMON(int)  		checkBags(void);
+COMMON(void)		markAtomsFindall(PL_local_data_t *ld);
 
 /* pl-comp.c */
 COMMON(void) 		initWamTable(void);
@@ -174,6 +172,7 @@ COMMON(code) 		replacedBreak(Code PC);
 COMMON(void) 		clearBreakPointsClause(Clause clause);
 COMMON(word) 		pl_current_break(term_t ref, term_t pc, control_t h);
 COMMON(int) 		get_clause_ptr_ex(term_t ref, Clause *cl);
+COMMON(int)		unify_functor(term_t t, functor_t fd, int how);
 
 /* pl-dump.c */
 COMMON(word) 		saveProgram(term_t new);
@@ -239,6 +238,7 @@ COMMON(int) 		PL_get_list_ex(term_t l, term_t h, term_t t);
 COMMON(int) 		PL_get_nil_ex(term_t l);
 COMMON(int) 		PL_unify_bool_ex(term_t t, bool val);
 COMMON(int) 		PL_get_arg_ex(int n, term_t term, term_t arg);
+COMMON(int) 		PL_get_module_ex(term_t name, Module *m);
 
 
 /* pl-file.c */
@@ -251,7 +251,7 @@ COMMON(bool) 		getInputStream__LD(term_t t, IOSTREAM **s ARG_LD);
 COMMON(bool) 		getOutputStream(term_t t, IOSTREAM **s);
 COMMON(bool) 		streamStatus(IOSTREAM *s);
 COMMON(atom_t) 		fileNameStream(IOSTREAM *s);
-COMMON(int) 		getSingleChar(IOSTREAM *s);
+COMMON(int) 		getSingleChar(IOSTREAM *s, int signals);
 COMMON(bool) 		readLine(IOSTREAM *in, IOSTREAM *out, char *buffer);
 COMMON(bool) 		LockStream(void);
 COMMON(bool) 		UnlockStream(void);
@@ -373,6 +373,7 @@ COMMON(term_t) 		PL_new_term_ref__LD(ARG1_LD);
 COMMON(term_t) 		PL_new_term_refs__LD(int n ARG_LD);
 COMMON(int) 		PL_unify__LD(term_t t1, term_t t2 ARG_LD);
 COMMON(int) 		PL_unify_integer__LD(term_t t1, intptr_t i ARG_LD);
+COMMON(int)		PL_unify_int64_ex__LD(term_t t1, int64_t ARG_LD);
 COMMON(int) 		PL_get_atom__LD(term_t t1, atom_t *a ARG_LD);
 COMMON(void) 		PL_put_atom__LD(term_t t1, atom_t a ARG_LD);
 COMMON(void) 		PL_put_integer__LD(term_t t1, long i ARG_LD);
@@ -460,6 +461,7 @@ COMMON(word) 		pl_open_shared_object(term_t file, term_t h, term_t flags);
 COMMON(word) 		pl_close_shared_object(term_t plhandle);
 COMMON(word) 		pl_call_shared_object_function(term_t plhandle, term_t name);
 COMMON(word) 		pl_load_shared_object(term_t file, term_t entry);
+COMMON(void)		cleanupForeign(void);
 
 /* pl-modul.c */
 COMMON(Module) 		lookupModule(atom_t name);
@@ -474,7 +476,6 @@ COMMON(word) 		pl_module(term_t old, term_t new);
 COMMON(word) 		pl_set_source_module(term_t old, term_t new);
 COMMON(word) 		pl_declare_module(term_t name, term_t file, term_t line);
 COMMON(word) 		pl_export_list(term_t modulename, term_t list);
-COMMON(word) 		pl_export(term_t head);
 COMMON(word) 		pl_check_export(void);
 COMMON(word) 		pl_context_module(term_t module);
 COMMON(word) 		pl_import(term_t pred);
@@ -543,30 +544,18 @@ COMMON(bool) 		raw_unify_ptrs(Word t1, Word t2 ARG_LD);
 COMMON(bool) 		unify_ptrs(Word t1, Word t2 ARG_LD);
 COMMON(bool) 		can_unify(Word t1, Word t2);
 COMMON(int) 		compareStandard(Word t1, Word t2, int eq ARG_LD);
-COMMON(int) 		lengthList(term_t list, int errors);
-COMMON(word) 		pl_univ(term_t t, term_t l);
+COMMON(intptr_t)	skip_list(Word l, Word *tailp ARG_LD);
+COMMON(intptr_t) 	lengthList(term_t list, int errors);
 COMMON(int) 		numberVars(term_t t, nv_options *opts, int n ARG_LD);
 COMMON(word) 		pl_e_free_variables(term_t t, term_t l);
 COMMON(word) 		stringToList(char *s);
-COMMON(word) 		pl_atom_length(term_t w, term_t n);
-COMMON(word) 		pl_name(term_t atom, term_t string);
-COMMON(word) 		pl_atom_chars(term_t atom, term_t string);
-COMMON(word) 		pl_atom_codes(term_t atom, term_t string);
-COMMON(word) 		pl_number_chars(term_t number, term_t string);
-COMMON(word) 		pl_number_codes(term_t number, term_t string);
-COMMON(word) 		pl_char_code(term_t atom, term_t chr);
 COMMON(word) 		pl_atom_concat(term_t a1, term_t a2, term_t a3,
 				       control_t ctx);
-COMMON(word) 		pl_concat_atom(term_t list, term_t atom);
-COMMON(word) 		pl_concat_atom3(term_t list, term_t sep, term_t atom);
 COMMON(word) 		pl_apropos_match(term_t a1, term_t a2);
 COMMON(foreign_t) 	pl_sub_atom(term_t atom,
 				    term_t before, term_t len, term_t after,
 				    term_t sub, control_t h);
-COMMON(word) 		pl_string_length(term_t str, term_t l);
 COMMON(word) 		pl_string_concat(term_t a1, term_t a2, term_t a3, control_t h);
-COMMON(word) 		pl_string_to_atom(term_t str, term_t a);
-COMMON(word) 		pl_string_to_list(term_t str, term_t list);
 COMMON(word) 		pl_sub_string(term_t str,
 			      term_t offset, term_t length, term_t after,
 			      term_t sub, control_t h);
@@ -614,7 +603,10 @@ COMMON(ClauseRef) 	hasClausesDefinition(Definition def);
 COMMON(bool) 		isDefinedProcedure(Procedure proc);
 COMMON(int) 		get_head_functor(term_t head, functor_t *fdef,
 				 int flags ARG_LD);
-COMMON(int) 		get_procedure(term_t descr, Procedure *proc, term_t he, int f);
+COMMON(int)		get_functor(term_t descr, functor_t *fdef,
+				    Module *m, term_t h, int how);
+COMMON(int) 		get_procedure(term_t descr, Procedure *proc,
+				      term_t he, int f);
 COMMON(word) 		pl_current_predicate(term_t name, term_t functor, control_t h);
 COMMON(foreign_t) 	pl_current_predicate1(term_t spec, control_t ctx);
 COMMON(ClauseRef) 	assertProcedure(Procedure proc, Clause clause,
@@ -638,23 +630,22 @@ COMMON(word) 		pl_get_clause_attribute(term_t ref, term_t att, term_t value);
 COMMON(word) 		pl_get_predicate_attribute(term_t pred, term_t k, term_t v);
 COMMON(word) 		pl_set_predicate_attribute(term_t pred, term_t k, term_t v);
 COMMON(void) 		reindexDefinition(Definition def);
-COMMON(void) 		redefineProcedure(Procedure proc, SourceFile sf);
+COMMON(void) 		redefineProcedure(Procedure proc, SourceFile sf,
+					  unsigned int suppress);
 COMMON(void) 		startConsult(SourceFile f);
 COMMON(void) 		indexDefinition(Definition def, long pattern);
 COMMON(word) 		pl_index(term_t pred);
-COMMON(SourceFile) 	lookupSourceFile(atom_t name);
+COMMON(SourceFile) 	lookupSourceFile(atom_t name, int create);
 COMMON(SourceFile) 	indexToSourceFile(int index);
 COMMON(void) 		cleanupSourceFiles(void);
 COMMON(void) 		addProcedureSourceFile(SourceFile sf, Procedure proc);
 COMMON(word) 		pl_make_system_source_files(void);
 COMMON(word) 		pl_source_file(term_t descr, term_t file, control_t h);
-COMMON(word) 		pl_time_source_file(term_t file, term_t t, control_t h);
 COMMON(word) 		pl_start_consult(term_t file);
 COMMON(word) 		pl_default_predicate(term_t d1, term_t d2);
 COMMON(Definition) 	autoImport(functor_t f, Module m);
 COMMON(word) 		pl_require(term_t pred);
 COMMON(word) 		pl_check_definition(term_t spec);
-COMMON(word) 		pl_clause_from_source(term_t file, term_t line, term_t clause);
 COMMON(foreign_t) 	pl_list_generations(term_t desc);
 COMMON(foreign_t) 	pl_check_procedure(term_t desc);
 COMMON(void) 		checkDefinition(Definition def);
@@ -672,6 +663,7 @@ COMMON(void) 		profRedo(struct call_node *node ARG_LD);
 
 /* pl-read.c */
 COMMON(void) 		resetRead(void);
+COMMON(int)		unicode_separator(pl_wchar_t c);
 COMMON(int) 		str_number(const unsigned char *string,
 				   unsigned char **end,
 				   Number value, bool escape);
@@ -681,7 +673,6 @@ COMMON(word) 		pl_read(term_t term);
 COMMON(word) 		pl_read2(term_t stream, term_t term);
 COMMON(word) 		pl_read_term(term_t term, term_t pos);
 COMMON(word) 		pl_read_term3(term_t stream, term_t term, term_t pos);
-COMMON(word) 		pl_atom_to_term(term_t term, term_t atom, term_t bindings);
 COMMON(void) 		initCharConversion(void);
 COMMON(foreign_t) 	pl_char_conversion(term_t in, term_t out);
 COMMON(foreign_t) 	pl_current_char_conversion(term_t in, term_t out, control_t h);
@@ -702,6 +693,7 @@ COMMON(word) 		pl_recorded(term_t key, term_t term, term_t ref, control_t h);
 COMMON(word) 		pl_erase(term_t ref);
 COMMON(word) 		pl_term_complexity(term_t t, term_t mx, term_t count);
 COMMON(void) 		undo_while_saving_term(mark *m, Word term);
+COMMON(void)		markAtomsRecord(Record record);
 
 /* pl-rl.c */
 COMMON(void) 		install_rl(void);
@@ -728,7 +720,7 @@ COMMON(bool) 		restoreStack(Stack s);
 COMMON(void) 		trimStacks(ARG1_LD);
 COMMON(void) 		resetStacks(void);
 COMMON(void) 		emptyStacks(void);
-COMMON(void) 		freeStacks(PL_local_data_t *ld);
+COMMON(void) 		freeStacks(ARG1_LD);
 COMMON(void) 		freeLocalData(PL_local_data_t *ld);
 COMMON(word) 		pl_trim_stacks(void);
 COMMON(word) 		pl_limit_stack(term_t s, term_t l);
@@ -920,8 +912,6 @@ COMMON(foreign_t) 	pl_open_xterm(term_t title, term_t in, term_t out, term_t err
 
 /* pl-ctype.c */
 
-COMMON(foreign_t) 	pl_downcase_atom(term_t in, term_t out);
-COMMON(foreign_t) 	pl_upcase_atom(term_t in, term_t out);
 COMMON(IOENC)		initEncoding(void);
 COMMON(void) 		initCharTypes(void);
 COMMON(bool) 		systemMode(bool accept);
@@ -936,3 +926,14 @@ COMMON(int)		unify_thread_id(term_t id, PL_thread_info_t *info);
 COMMON(int) 		enableThreads(int enable);
 
 
+/* pl-gmp.c */
+COMMON(int)	PL_unify_number(term_t t, Number n);
+COMMON(void)	get_number(word w, Number n  ARG_LD);
+COMMON(int)	PL_get_number(term_t t, Number n);
+COMMON(word)	put_number(Number n);
+COMMON(void)	promoteToRealNumber(Number n);
+COMMON(void)	same_type_numbers(Number n1, Number n2);
+COMMON(void)    promoteNumber(Number n1, numtype type);
+COMMON(int)	cmpNumbers(Number n1, Number n2);
+COMMON(void)	cpNumber(Number to, Number from);
+
