@@ -1,4 +1,4 @@
-/*  $Id: pl-setup.c,v 1.171 2007/06/18 17:44:32 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -347,6 +347,7 @@ dispatch_signal(int sig, int sync)
   fid_t fid;
   uintptr_t lTopSave;
   int saved_current_signal;
+  int saved_sync;
 
 #ifdef O_PLMT
   if ( !LD )
@@ -355,15 +356,19 @@ dispatch_signal(int sig, int sync)
 
     return;				/* what else?? */
   }
-#endif
 
   DEBUG(1, Sdprintf("Got signal %d in thread %d (=%d) %s\n",
 		    sig, LD->thread.info->pl_tid,
 		    pthread_self(),
 		    sync ? " (sync)" : " (async)"));
-	
+#else
+  DEBUG(1, Sdprintf("Got signal %d %s\n",
+		    sig, sync ? " (sync)" : " (async)"));
+#endif
+
   lTopSave = (char*)lTop - (char*)lBase;
   saved_current_signal = LD->current_signal;
+  saved_sync = LD->sync_signal;
 
   switch(sig)
   { case SIGFPE:
@@ -387,6 +392,7 @@ dispatch_signal(int sig, int sync)
 
   blockGC(PASS_LD1);
   LD->current_signal = sig;
+  LD->sync_signal = sync;
   fid = PL_open_signal_foreign_frame();
 
   DEBUG(1, Sdprintf("Handling signal %d, pred = %p, handler = %p\n",
@@ -437,6 +443,7 @@ dispatch_signal(int sig, int sync)
   }
 
   LD->current_signal = saved_current_signal;
+  LD->sync_signal = saved_sync;
   PL_discard_foreign_frame(fid);
   lTop = addPointer(lBase, lTopSave);
 
@@ -724,6 +731,9 @@ PL_signal(int sigandflags, handler_t func)
 }
 
 
+/* return: -1: exception in handler, otherwise number of handled signals
+*/
+
 int
 PL_handle_signals()
 { int done = 0;
@@ -776,6 +786,12 @@ PL_handle_signals()
 
   if ( exception_term )
     return -1;
+
+  if ( gc_status.requested )
+  { DEBUG(1, Sprintf("GC from PL_handle_signals()\n"));
+    garbageCollect(NULL, NULL);
+  }
+
   return done;
 }
 
@@ -905,6 +921,8 @@ pl-thread.c to create stacks for Prolog threads.
 
 allocStacks() does the  real  work   and  has  several  implementations,
 depending on the OS features.
+
+Requested stack sizes are in bytes.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
@@ -1180,7 +1198,9 @@ unmap(Stack s)
   { intptr_t len = (char *)s->max - (char *)addr;
 
 #ifdef MAP_FIXED
+#ifndef MMAP_OVERRIDES
     munmap(addr, len);
+#endif
     if ( mmap(addr, len, MAPPROTECT, MAP_FIXED|MAP_FLAGS, mapfd, 0L) != addr )
       fatalError("Failed to remap 0x%x bytes at %p: %s",
 		 len, addr, OsError());
@@ -1660,7 +1680,7 @@ allocStacks(intptr_t local, intptr_t global, intptr_t trail, intptr_t argument)
   tBase = (TrailEntry) malloc(itrail);
   aBase = (Word *)     malloc(argument);
   if ( !gBase || !tBase || !aBase )
-  { freeStacks(LD);
+  { freeStacks(PASS_LD1);
     fail;
   }
 
@@ -1731,9 +1751,6 @@ trimStacks(ARG1_LD)
 #endif /*O_DYNAMIC_STACKS*/
 #endif /*O_SHIFT_STACKS*/
 
-  LD->stacks.global.gced_size = usedStack(global);
-  LD->stacks.trail.gced_size  = usedStack(trail);
-
   for(te = tTop; --te >= tBase; )
   { Word p = te->address;
     
