@@ -1,4 +1,4 @@
-/*  $Id: pl-trace.c,v 1.116 2007/06/01 14:24:32 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -401,7 +401,7 @@ again:
 	exitFromDebugger(0);
       }
     } else
-    { int c = getSingleChar(Sdin);
+    { int c = getSingleChar(Sdin, FALSE);
 
       if ( c == EOF )
       { Sfputs("EOF: exit\n", Sdout);
@@ -579,6 +579,7 @@ traceAction(char *cmd, int port, LocalFrame frame, Choice bfr, bool interactive)
     case EOS:
     case ' ':
     case '\n':
+    case '\r':
     case 'c':	FeedBack("creep\n");
 		clear(frame, FR_SKIPPED);
 		return ACTION_CONTINUE;
@@ -1087,7 +1088,7 @@ traceInterception(LocalFrame frame, Choice bfr, int port, Code PC)
 #endif /*O_DEBUGGER*/
 
 #ifndef offset
-#define offset(s, f) ((int)(&((struct s *)NULL)->f))
+#define offset(s, f) ((size_t)(&((struct s *)NULL)->f))
 #endif
 
 static QueryFrame
@@ -1096,7 +1097,7 @@ findQuery(LocalFrame fr)
     fr = fr->parent;
 
   if ( fr )
-    return (QueryFrame)addPointer(fr, -offset(queryFrame, frame));
+    return QueryOfTopFrame(fr);
   return NULL;
 }
 
@@ -1208,7 +1209,7 @@ resetTracer(void)
   debugstatus.skiplevel    = 0;
   debugstatus.retryFrame   = NULL;
 
-  setFeatureMask(TAILRECURSION_FEATURE);
+  setFeatureMask(LASTCALL_FEATURE);
 }
 
 
@@ -1236,13 +1237,13 @@ helpInterrupt(void)
 static void
 interruptHandler(int sig)
 { int c; 
+  int safe;
 
   if ( !GD->initialised )
   { Sfprintf(Serror, "Interrupt during startup. Cannot continue\n");
     PL_halt(1);
   }  
 
-again:
 #ifdef O_PLMT
   if ( !LD )				/* we can't handle this; main thread */
   { PL_thread_raise(1, sig);		/* should try to do this */
@@ -1257,10 +1258,29 @@ again:
   }
 #endif
 
-  Sfputs("\nAction (h for help) ? ", Sdout);
+#if __unix__				/* actually, asynchronous signal handling */
+  if ( !LD->sync_signal )
+  { long mask = (1L<<(sig-1));
+
+    if ( (LD->pending_signals & mask) )
+    { LD->pending_signals &= ~mask;
+      safe = FALSE;
+    } else
+    { DEBUG(1, Sdprintf("Reposting as synchronous\n"));
+      PL_raise(sig);
+      return;
+    }
+  } else
+  { safe = TRUE;
+  }
+#endif
+
+  Sreset();
+again:
+  Sfprintf(Sdout, "\n%sAction (h for help) ? ", safe ? "" : "[forced] ");
   Sflush(Sdout);
   ResetTty();                           /* clear pending input -- atoenne -- */
-  c = getSingleChar(Sdin);
+  c = getSingleChar(Sdin, FALSE);
 
   switch(c)
   { case 'a':	Sfputs("abort\n", Sdout);
@@ -1366,7 +1386,7 @@ debugmode(debug_type doit, debug_type *old)
   if ( debugstatus.debugging != doit )
   { if ( doit )
     { debugstatus.skiplevel = VERY_DEEP;
-      clearFeatureMask(TAILRECURSION_FEATURE);
+      clearFeatureMask(LASTCALL_FEATURE);
       if ( doit == DBG_ALL )
       { LocalFrame fr = environment_frame;
 
@@ -1374,8 +1394,7 @@ debugmode(debug_type doit, debug_type *old)
 	{ if ( fr->parent )
 	    fr = fr->parent;
 	  else
-	  { QueryFrame qf = (QueryFrame)addPointer(fr, -offset(queryFrame,
-							       frame));
+	  { QueryFrame qf = QueryOfTopFrame(fr);
 	    assert(qf->magic == QID_MAGIC);
 	    qf->debugSave = DBG_ON;
 	    fr = qf->saved_environment;
@@ -1384,7 +1403,7 @@ debugmode(debug_type doit, debug_type *old)
 	doit = DBG_ON;
       }
     } else
-    { setFeatureMask(TAILRECURSION_FEATURE);
+    { setFeatureMask(LASTCALL_FEATURE);
     }
     debugstatus.debugging = doit;
     printMessage(ATOM_silent,
