@@ -1,11 +1,11 @@
-/*  $Id: test.pl,v 1.150 2007/06/19 09:05:32 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker and Anjo Anjewierden
     E-mail:        wielemak@science.uva.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2005, University of Amsterdam
+    Copyright (C): 1985-2007, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -26,7 +26,7 @@
 %:- set_prolog_flag(trace_gc, true).
 :- asserta(user:file_search_path(library, '../packages/plunit')).
 :- [library(plunit)].
-:- set_test_options([load(always), silent(true)]).
+:- set_test_options([load(always), silent(true), sto(true)]).
 
 :- use_module(library(lists)).
 
@@ -146,7 +146,11 @@ syntax(number-2) :-
 syntax(zero-1) :-
 	term_to_atom(T, 'hello(\000\"\000\x")'),
 	T == hello([0, 120]).
-
+syntax(latin-1) :-
+	atom_codes(A, [247]),
+	atom_to_term(A, T, []),
+	atom_codes(T, [247]).
+	
 
 		 /*******************************
 		 *	       WRITE		*
@@ -159,11 +163,18 @@ write_test(q-2) :-
 write_test(q-3) :-
 	term_to_atom(+(a), X), X == '+a'.
 write_test(q-4) :-
-	term_to_atom('/*', X), X == '\'/*\''.
+	term_to_atom('/*', X), X == '\'/*\''.	%0'
 write_test(q-5) :-
 	term_to_atom('/**', X), X == '\'/**\''.
 write_test(q-6) :-
 	term_to_atom('*/*', X), X == '*/*'.
+write_test(q-7) :-
+	term_to_atom(p(0|a), X), X == 'p((0 \'|\' a))'.
+write_test(q-8) :-
+	term_to_atom(p(a|b), X), X == 'p((a\'|\'b))'.
+write_test(c-1) :-
+	T = [a,b,c|T],
+	term_to_atom(T, X), X == '[a, b, c|**]'.
 
 
 		 /*******************************
@@ -245,13 +256,26 @@ cyclic(hash_term-1) :-
 cyclic(streq-1) :-
 	X = [X], Y = [Y], X =@= Y.
 cyclic(test-1) :-
-	X = f(X), cyclic_term(X).
+	X = f(X),
+	cyclic_term(X).
 cyclic(test-2) :-
-	acyclic_term(f(x)).
+	X = f(X),
+	cyclic_term(a(X, a)).
 cyclic(test-3) :-
-	acyclic_term(_).
+	X = f(X),
+	cyclic_term(a(a, X)).
 cyclic(test-4) :-
+	acyclic_term(f(x)).
+cyclic(test-5) :-
+	acyclic_term(_).
+cyclic(test-6) :-
 	X = f(a), acyclic_term(a(X, X)).
+cyclic(list-1) :-
+	L = [a|L], \+ is_list(L).
+cyclic(sort-1) :-
+	L = [a,b,c|L],
+	sort(L, List),
+	List == [a,b,c].
 
 
 		 /*******************************
@@ -574,8 +598,12 @@ gmp(pow-2) :-
 gmp(pow-3) :-
 	A is 0**0, A = 1.
 gmp(pow-4) :-
-	A is 1.5**2, A = 2.25.
+	A is -1**0, A = 1.
 gmp(pow-5) :-
+	A is -1**0, A = 1.
+gmp(pow-6) :-
+	A is -1**((1<<100)+1), A == -1.
+gmp(pow-7) :-
 	A is -100**5,
 	A < 0,
 	abs(A) =:= 100**5.
@@ -837,7 +865,9 @@ cleanup(clean-6) :-
 cleanup(clean-7) :-
 	catch(call_cleanup(fail, throw(b)), E, true),
 	E == b.
-cleanup(clean-8) :-
+cleanup(clean-8) :-			% check handling of CHP_TOP
+	'$break'(call_cleanup(true, true)).
+cleanup(clean-9) :-
 	retractall(clean_rval(_)),
 	call_cleanup(bagof(x, cleanup_1, _Xs), Reason,
 		     assert(clean_rval(Reason))),
@@ -938,6 +968,10 @@ term(univ-3) :-
 	3.4 =.. X, X == [3.4].
 term(univ-4) :-
 	a(a,b,c) =.. [a, a | L], L == [b,c].
+term(univ-5) :-
+	L = [a|L],
+	catch(_T =.. L, error(E, _), true),
+	E == type_error(list, L).
 
 
 
@@ -993,19 +1027,34 @@ set(X, 1) :- type(X).
 set(X, 2) :- type(X).
 
 sets(setof-1) :-
-	setof(A-Pairs, setof(B, foo(A,B), Pairs), Result),
-	Result = [1 - [a,d],2 - [b,e],3 - [c,f]].
+	setof(A-Pairs, setof(B, foo(A,B), Pairs), Result0),
+	keysort(Result0, Result),
+	Result = [ 1 - [a,d],
+		   2 - [b,e],
+		   3 - [c,f]
+		 ].
 sets(setof-2) :-
-	setof(X-Ys, setof(Y, set(X,Y), Ys), R),
+	setof(X-Ys, setof(Y, set(X,Y), Ys), R0),
 	string_to_atom(S, "string"),
-	R =@= [3.14-[1, 2],
-	       42-[1, 2],
-	       atom-[1, 2],
-	       S-[1, 2],
-	       compound(1)-[1, 2],
-	       [a, list]-[1, 2],
-	       compound(_A, _B)-[1, 2],
-	       compound(A, A)-[1, 2]].
+	keysort(R0, R),
+	(   R =@= [3.14-[1, 2],
+		   42-[1, 2],
+		   atom-[1, 2],
+		   S-[1, 2],
+		   compound(1)-[1, 2],
+		   [a, list]-[1, 2],
+		   compound(_A0, _B0)-[1, 2], % order is not defined
+		   compound(A, A)-[1, 2]]
+	->  true
+	;   R =@= [3.14-[1, 2],
+		   42-[1, 2],
+		   atom-[1, 2],
+		   S-[1, 2],
+		   compound(1)-[1, 2],
+		   [a, list]-[1, 2],
+		   compound(A, A)-[1, 2],
+		   compound(_A1, _B1)-[1, 2]]
+	).
 sets(vars-1) :-
 	'$e_free_variables'(A^satisfy(B^C^(setof(D:E,
 						 (country(E), area(E, D)),
@@ -1015,6 +1064,16 @@ sets(vars-1) :-
 					   {place(A)})),
 			    Free),
 	Free == v(D, E).
+sets(bagof-1) :-
+	List = [_,_,_],
+	bagof(X, member(X, List), Xs),
+	Xs == List.
+sets(setof-3) :-
+	List = [_,_,_],
+	setof(X, member(X, List), Xs),
+	sort(List, LS),
+	Xs == LS.
+
 
 		 /*******************************
 		 *	       NAME		*
@@ -1122,6 +1181,12 @@ proc(retract-2) :-
 	assert((test(X, Y) :- X is Y + 3)),
 	retract((test(A, B) :- Body)),
 	Body == (A is B + 3).
+proc(retract-3) :-
+	assert(myunit(1)),
+	assert(myunit(2) :- x),
+	retract(myunit(2) :- X),
+	X == x,
+	retractall(myunit(_)).		% cleanup
 proc(current_predicate-1) :-
 	setof(X, current_predicate(cpxx/X), L), % order is not defined!
 	L == [0, 2].
@@ -1131,6 +1196,9 @@ proc(erase-static) :-
 	clause(cpxx(_,_), true, Ref),
 	catch(erase(Ref), E, true),
 	nonvar(E).
+proc(current_predicate-2) :-
+	catch(current_predicate(foo(bar) : baz/0), E, true),
+	E = error(type_error(atom, foo(bar)), _).
 
 cp_one :-
 	assert(cp_foo(a)),
@@ -1559,6 +1627,12 @@ gc(gc-3) :-
 		garbage_collect,
 		X == a
 	      ).
+gc(gc-4) :-
+	catch(_,_,garbage_collect).
+gc(gc-5) :-
+	catch(25,_,garbage_collect).
+gc(gc-6) :-
+	catch(1.25,_,garbage_collect).
 gc(agc-1) :-
 	garbage_collect_atoms.
 gc(agc-2) :-
@@ -2237,6 +2311,8 @@ ctype(code_type-4) :-
 ctype(code_type-5) :-
 	code_type(48, digit(W)),
 	W == 0.
+ctype(code_type-6) :-
+	\+ code_type(0, quote).
 
 		 /*******************************
 		 *	       WCTYPE		*
@@ -2313,7 +2389,7 @@ thread(status-1) :-
 	thread_create(true, Id, []),
 	between(0, 10, _),
 	sleep(0.01),
-	current_thread(Id2, Status),
+	thread_property(Id2, status(Status)),
 	Id2 == Id,
 	Status == true, !,
 	thread_join(Id2, _).
@@ -2509,6 +2585,7 @@ wide_character_types :-
 %	
 %	Enumerate directories holding tests.
 
+testdir('Tests/core').
 testdir('Tests/attvar').
 testdir('Tests/library').
 testdir('Tests/charset').
